# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import numpy as np
import pandas as pd
from numpy import complex128
from pandapower.auxiliary import _sum_by_group, sequence_to_phase, _sum_by_group_nvals
from pandapower.pypower.idx_bus import VM, VA, PD, QD, LAM_P, LAM_Q, BASE_KV, NONE, BS, BUS_TYPE, BUS_I
from pandapower.pypower.idx_bus_dc import DC_VM, DC_BUS_TYPE, DC_NONE, DC_PD, DC_BUS_I

from pandapower.pypower.idx_gen import PG, QG
from pandapower.build_bus import _get_motor_pq, _get_symmetric_pq_of_unsymetric_element
from pandapower.pypower.idx_ssc import SSC_X_CONTROL_VM, SSC_X_CONTROL_VA, SSC_Q, SSC_INTERNAL_BUS
from pandapower.pypower.idx_svc import SVC_THYRISTOR_FIRING_ANGLE, SVC_Q, SVC_X_PU
from pandapower.pypower.idx_vsc import VSC_Q, VSC_P, VSC_P_DC, VSC_BUS_DC, VSC_INTERNAL_BUS_DC, VSC_INTERNAL_BUS

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def _set_buses_out_of_service(ppc):
    disco = np.where(ppc["bus"][:, BUS_TYPE] == NONE)[BUS_I]
    ppc["bus"][disco, VM] = np.nan
    ppc["bus"][disco, VA] = np.nan
    ppc["bus"][disco, PD] = 0
    ppc["bus"][disco, QD] = 0


def _set_dc_buses_out_of_service(ppc):
    disco = np.where(ppc["bus_dc"][:, DC_BUS_TYPE] == DC_NONE)[DC_BUS_I]
    ppc["bus_dc"][disco, DC_VM] = np.nan
    ppc["bus_dc"][disco, DC_PD] = 0


def _get_bus_v_results(net, ppc, suffix=None):
    bus_idx = _get_bus_idx(net)

    res_table = "res_bus" if suffix is None else "res_bus%s" % suffix
    # always read the vm_pu from ppc, even for DC power flow - otherwise old values can persist
    net[res_table]["vm_pu"] = ppc["bus"][bus_idx][:, VM]
    # voltage angles
    net[res_table]["va_degree"] = ppc["bus"][bus_idx][:, VA]


def _get_bus_dc_v_results(net, ppc):
    bus_idx = _get_bus_idx(net, "bus_dc")
    net["res_bus_dc"]["vm_pu"] = ppc["bus_dc"][bus_idx][:, DC_VM]


def _get_bus_v_results_3ph(net, ppc0, ppc1, ppc2):
    ac = net["_options"]["ac"]
    V012_pu = _V012_from_ppc012(net, ppc0, ppc1, ppc2)
    # Uncomment for results in kV instead of pu
    # bus_base_kv = ppc0["bus"][:,BASE_KV]/np.sqrt(3)
    # V012_pu = V012_pu*bus_base_kv

    Vabc_pu = sequence_to_phase(V012_pu)

    if ac:
        net["res_bus_3ph"]["vm_a_pu"] = np.abs(Vabc_pu[0, :].flatten())
        net["res_bus_3ph"]["vm_b_pu"] = np.abs(Vabc_pu[1, :].flatten())
        net["res_bus_3ph"]["vm_c_pu"] = np.abs(Vabc_pu[2, :].flatten())
    # voltage angles
    net["res_bus_3ph"]["va_a_degree"] = np.angle(Vabc_pu[0, :].flatten())*180/np.pi
    net["res_bus_3ph"]["va_b_degree"] = np.angle(Vabc_pu[1, :].flatten())*180/np.pi
    net["res_bus_3ph"]["va_c_degree"] = np.angle(Vabc_pu[2, :].flatten())*180/np.pi
    net["res_bus_3ph"]["unbalance_percent"] = np.abs(V012_pu[2, :]/V012_pu[1, :])*100
    net["res_bus_3ph"].index = net["bus"].index


def _V012_from_ppc012(net, ppc0, ppc1, ppc2):
    bus_idx = _get_bus_idx(net)
    V012_pu = np.zeros((3, len(bus_idx)), dtype=complex128)
    V012_pu[0, :] = ppc0["bus"][bus_idx][:, VM] * np.exp(1j * np.deg2rad(ppc0["bus"][bus_idx][:, VA]))
    V012_pu[1, :] = ppc1["bus"][bus_idx][:, VM] * np.exp(1j * np.deg2rad(ppc1["bus"][bus_idx][:, VA]))
    V012_pu[2, :] = ppc2["bus"][bus_idx][:, VM] * np.exp(1j * np.deg2rad(ppc2["bus"][bus_idx][:, VA]))
    return V012_pu


def _get_bus_idx(net, bus_table="bus"):
    bus_lookup = net["_pd2ppc_lookups"][bus_table]
    ppi = net[bus_table].index.values
    bus_idx = bus_lookup[ppi]
    return bus_idx


def _get_opf_marginal_prices(net, ppc):
    bus_idx = _get_bus_idx(net)
    net["res_bus"]["lam_p"] = ppc["bus"][bus_idx][:, LAM_P]
    net["res_bus"]["lam_q"] = ppc["bus"][bus_idx][:, LAM_Q]


def _get_bus_results(net, ppc, bus_pq):
    ac = net["_options"]["ac"]
    mode = net["_options"]["mode"]

    # write sum of p and q values to bus
    net["res_bus"]["p_mw"].values[:] = bus_pq[:, 0]
    if ac:
        net["res_bus"]["q_mvar"].values[:] = bus_pq[:, 1]

    # opf variables
    if mode == "opf":
        _get_opf_marginal_prices(net, ppc)

    # update index in res_bus
    net["res_bus"].index = net["bus"].index


def _get_bus_dc_results(net, bus_p_dc):

    # write sum of p and q values to bus
    net["res_bus_dc"]["p_mw"].values[:] = bus_p_dc[:, 0]

    # update index in res_bus_dc
    net["res_bus_dc"].index = net["bus_dc"].index


def _get_bus_results_3ph(net, bus_pq):
    ac = net["_options"]["ac"]

    # write sum of p and q values to bus
    net["res_bus_3ph"]["p_a_mw"] = bus_pq[:, 0]
    net["res_bus_3ph"]["p_b_mw"] = bus_pq[:, 2]
    net["res_bus_3ph"]["p_c_mw"] = bus_pq[:, 4]
    if ac:
        net["res_bus_3ph"]["q_a_mvar"] = bus_pq[:, 1]
        net["res_bus_3ph"]["q_b_mvar"] = bus_pq[:, 3]
        net["res_bus_3ph"]["q_c_mvar"] = bus_pq[:, 5]

    # Todo: OPF

    # update index in res bus bus
    # net["res_bus"].index = net["bus"].index
    net["res_bus_3ph"].index = net["bus"].index


def write_voltage_dependend_load_results(net, p, q, b):
    l = net["load"]
    _is_elements = net["_is_elements"]

    if len(l) > 0:
        load_is = _is_elements["load"]
        scaling = l["scaling"].values
        bus_lookup = net["_pd2ppc_lookups"]["bus"]
        lidx = bus_lookup[l["bus"].values]

        voltage_depend_loads = net["_options"]["voltage_depend_loads"]

        cz = l["const_z_percent"].values / 100.
        ci = l["const_i_percent"].values / 100.
        cp = 1. - (cz + ci)

        # constant power
        pl = l["p_mw"].values * scaling * load_is * cp
        net["res_load"]["p_mw"] = pl
        p = np.hstack([p, pl])

        ql = l["q_mvar"].values * scaling * load_is * cp
        net["res_load"]["q_mvar"] = ql
        q = np.hstack([q, ql])

        b = np.hstack([b, l["bus"].values])

        if voltage_depend_loads:
            # constant impedance and constant current
            vm_l = net["_ppc"]["bus"][lidx, 7]
            volt_depend = ci * vm_l + cz * vm_l ** 2
            pl = l["p_mw"].values * scaling * load_is * volt_depend
            net["res_load"]["p_mw"] += pl
            p = np.hstack([p, pl])

            ql = l["q_mvar"].values * scaling * load_is * volt_depend
            net["res_load"]["q_mvar"] += ql
            q = np.hstack([q, ql])

            b = np.hstack([b, l["bus"].values])
        return p, q, b


def write_pq_results_to_element(net, ppc, element, suffix=None):
    """
    get p_mw and q_mvar for a specific pq element ("load", "sgen"...).
    This function basically writes values element table to res_element table
    :param net: pandapower net
    :param element: element name (str)
    :return:
    """
    # info from net
    _is_elements = net["_is_elements"]
    ac = net["_options"]["ac"]

    # info element
    el_data = net[element]
    res_ = "res_%s" % element
    if suffix is not None:
        res_ += "_%s" % suffix
    ctrl_ = "%s_controllable" % element

    is_controllable = False
    if ctrl_ in _is_elements:
        controlled_elements = net[element][net._is_elements[ctrl_]].index
        gen_idx = net._pd2ppc_lookups[ctrl_][controlled_elements]
        gen_sign = 1 if element == "sgen" else -1
        is_controllable = True

    if element == "motor":
        p_mw, q_mvar = _get_motor_pq(net)
        net[res_]["p_mw"].values[:] = p_mw
        net[res_]["q_mvar"].values[:] = q_mvar
        return net
    elif element.startswith("asymmetric"):
        p_mw, q_mvar = _get_symmetric_pq_of_unsymetric_element(net, element)
        net[res_]["p_mw"].values[:] = p_mw
        net[res_]["q_mvar"].values[:] = q_mvar
        return net

    # Wards and xwards have different names in their element table, but not in res table. Also no scaling -> Fix...
    p_mw = "ps_mw" if element in ["ward", "xward"] else "p_mw"
    q_mvar = "qs_mvar" if element in ["ward", "xward"] else "q_mvar"
    scaling = el_data["scaling"].values if element not in ["ward", "xward"] else 1.0

    element_in_service = _is_elements[element]

    # P result in mw to element
    net[res_]["p_mw"].values[:] = el_data[p_mw].values * scaling * element_in_service
    if is_controllable:
        net[res_].loc[controlled_elements, "p_mw"] = ppc["gen"][gen_idx, PG] * gen_sign

    # add results from distributed slack calculation for xwards from ppc instead of the element table
    if net._options['distributed_slack'] and element == "xward" and np.any(net.xward.loc[element_in_service, 'slack_weight'].values != 0):
        _extract_dist_slack_pq_results(net, ppc, element, res_)

    if ac:
        # Q result in mvar to element
        net[res_]["q_mvar"].values[:] = el_data[q_mvar].values * scaling * element_in_service
        if is_controllable:
            net[res_].loc[controlled_elements, "q_mvar"] = ppc["gen"][gen_idx, QG] * gen_sign
    else:
        net[res_]["q_mvar"].values[:] = np.nan

    return net


def write_p_dc_results_to_element(net, ppc, element):
    """
    get p_mw for a specific DC pq element ("vsc").
    This function basically writes values element table to res_element table
    :param net: pandapower net
    :param element: element name (str)
    :return:
    """
    # info from net
    _is_elements = net["_is_elements"]
    ac = net["_options"]["ac"]

    # info element
    element_data = net[element]
    res_ = "res_%s" % element

    # todo: how does "not controllable" element behave?
    ctrl_ = "%s_controllable" % element

    is_controllable = False
    if ctrl_ in _is_elements:
        raise UserWarning("wrong idea about controllable")
        controlled_elements = net[element][net._is_elements[ctrl_]].index
        gen_idx = net._pd2ppc_lookups[ctrl_][controlled_elements]
        gen_sign = 1 if element == "sgen" else -1
        is_controllable = True

    if element != "vsc":
        raise NotImplementedError("Only VSC DC element is implemented for bus_dc P results")

    p_mw = "control_value_dc"  # in "vsc" element
    vsc_p_mode = _is_elements[element] & (net.vsc.control_mode_dc == "p_mw")

    # P result in mw to element
    # net[res_]["p_dc_mw"].values[:] = element_data[p_mw].values * vsc_p_mode

    # use the ppc value for the result instead:
    #res_p = np.nans(shape=(len(net[element])), dtype=np.float64)
    #res_p[net._is_elements["vsc"]] = ppc[element][:, VSC_P_DC]
    net[res_]["p_dc_mw"].values[:] = ppc[element][:, VSC_P_DC]
    return net


def _extract_dist_slack_pq_results(net, ppc, element, res_):
    node_elements = ['sgen', 'load', 'ward', 'xward', 'storage']
    for b in net[element].bus.values:
        connected = dict()
        for e in node_elements:
            conn = net[e].loc[net[e].in_service & (net[e].bus == b)].index.values
            if len(conn) > 0:
                connected.update({e: conn})

        p_bus = ppc['bus'][net._pd2ppc_lookups["bus"][b], PD]
        # first, count the total slack weights per bus, and obtain the variable part of the active power
        total_weight = 0
        for e, idx in connected.items():
            if 'slack_weight' in net[e].columns:
                elm_weight = net[e].loc[idx, 'slack_weight'].values
                if np.abs(elm_weight).sum() != 0:
                    total_weight += np.abs(elm_weight)
            p_bus -= net[e].loc[idx, 'ps_mw' if e in ["ward", "xward"] else "p_mw"].values.sum()

        # now, distribute the variable part of the active power among the dist_slack elements
        for e, idx in connected.items():
            if 'slack_weight' in net[e].columns:
                elm_weight = net[e].loc[idx, 'slack_weight'].values
                if np.abs(elm_weight).sum() != 0:
                    net[res_]["p_mw"] += p_bus * elm_weight / total_weight


def write_pq_results_to_element_3ph(net, element):
    """
    get p_mw and q_mvar for a specific pq element ("load", "sgen"...).
    This function basically writes values element table to res_element table

    :param net: pandapower net
    :param element: element name (str)
    :return:
    """

    # info from net
    _is_elements = net["_is_elements"]
    ac = net["_options"]["ac"]

    # info element
    el_data = net[element]
    res_ = "res_" + element+"_3ph"

    scaling = el_data["scaling"].values

    element_in_service = _is_elements[element]

    net[res_]["p_a_mw"] = list(
        (el_data["p_mw"].values / 3) * scaling * element_in_service) if element in ["load", "sgen"] \
        else list(el_data["p_a_mw"].values * scaling * element_in_service)

    net[res_]["p_b_mw"] = list(
        (el_data["p_mw"].values / 3) * scaling * element_in_service)if element in ["load", "sgen"] \
        else list(el_data["p_b_mw"].values * scaling * element_in_service)

    net[res_]["p_c_mw"] = list(
        (el_data["p_mw"].values / 3) * scaling * element_in_service) if element in ["load", "sgen"] \
        else list(el_data["p_c_mw"].values * scaling * element_in_service)
    if ac:
        # Q result in kvar to element
        net[res_]["q_a_mvar"] = list(
            (el_data["q_mvar"].values / 3) * scaling * element_in_service) if element in ["load", "sgen"] \
            else list(el_data["q_a_mvar"].values * scaling * element_in_service)

        net[res_]["q_b_mvar"] = list(
            (el_data["q_mvar"].values / 3) * scaling * element_in_service) if element in ["load", "sgen"] \
            else list(el_data["q_b_mvar"].values * scaling * element_in_service)

        net[res_]["q_c_mvar"] = list(
            (el_data["q_mvar"].values / 3) * scaling * element_in_service) if element in ["load", "sgen"] \
            else list(el_data["q_c_mvar"].values * scaling * element_in_service)

    # update index of result table
    net[res_].index = net[element].index
    return net


def get_p_q_b(net, element, suffix=None):
    ac = net["_options"]["ac"]
    res_ = "res_" + element
    if suffix != None:
        res_ += "_%s" % suffix

    # bus values are needed for stacking
    b = net[element]["bus"].values
    p = net[res_]["p_mw"]
    q = net[res_]["q_mvar"] if ac else np.zeros_like(p)
    return p, q, b


def get_p_q_b_3ph(net, element):
    ac = net["_options"]["ac"]
    res_ = "res_" + element+"_3ph"

    # bus values are needed for stacking
    b = net[element]["bus"].values
    pA = net[res_]["p_a_mw"]
    pB = net[res_]["p_b_mw"]
    pC = net[res_]["p_c_mw"]
    qA = net[res_]["q_a_mvar"] if ac else np.zeros_like(pA)
    qB = net[res_]["q_b_mvar"] if ac else np.zeros_like(pB)
    qC = net[res_]["q_c_mvar"] if ac else np.zeros_like(pC)
    return pA, qA, pB, qB, pC, qC, b


def _get_p_q_results(net, ppc, bus_lookup_aranged):
    bus_pq = np.zeros(shape=(len(net["bus"].index), 2), dtype=np.float64)
    b, p, q = np.array([]), np.array([]), np.array([])

    ac = net["_options"]["ac"]
    elements = ["load", "motor", "sgen", "storage", "ward", "xward",
                "asymmetric_load", "asymmetric_sgen"]

    if net["_options"]["voltage_depend_loads"] and ac:
        # voltage dependend loads need special treatment here

        p, q, b = write_voltage_dependend_load_results(net, p, q, b)
        elements.remove("load")

    for element in elements:
        if len(net[element]) > 0:
            write_pq_results_to_element(net, ppc, element)
            p_el, q_el, bus_el = get_p_q_b(net, element)
            if element.endswith("sgen"):
                p = np.hstack([p, -p_el])
                q = np.hstack([q, -q_el])
            else:
                p = np.hstack([p, p_el])
                q = np.hstack([q, q_el])
            b = np.hstack([b, bus_el])

    if not ac:
        q = np.zeros(len(p))

    # sum pq results from every element to be written to net['bus'] later on
    b_pp, vp, vq = _sum_by_group(b.astype(np.int64), p, q)
    b_ppc = bus_lookup_aranged[b_pp]
    bus_pq[b_ppc, 0] = vp
    bus_pq[b_ppc, 1] = vq
    return bus_pq


def _get_p_dc_results(net, ppc, bus_lookup_aranged):
    bus_p_dc = np.zeros(shape=(len(net["bus_dc"].index), 1), dtype=np.float64)  # 1 because only p relevant
    b, p = np.array([]), np.array([])

    # ac = net["_options"]["ac"]
    # elements = ["load", "motor", "sgen", "storage", "ward", "xward",
    #             "asymmetric_load", "asymmetric_sgen"]
    elements = ["vsc"]  # we only have VSC element so far that injects or consumes P from DC bus

    for element in elements:
        if len(net[element]) > 0:
            write_p_dc_results_to_element(net, ppc, element)
            bus_el = net[element]["bus_dc"].values
            p_el = net[f"res_{element}"]["p_dc_mw"]
            # if element.endswith("sgen"):
            #     p = np.hstack([p, -p_el])
            # else:
            #     p = np.hstack([p, p_el])
            p = np.hstack([p, p_el])
            b = np.hstack([b, bus_el])

    # sum pq results from every element to be written to net['bus'] later on
    b_pp, vp, _ = _sum_by_group(b.astype(np.int64), p, p)
    b_ppc = bus_lookup_aranged[b_pp]
    bus_p_dc[b_ppc, 0] = vp
    return bus_p_dc


def _get_p_q_results_3ph(net, bus_lookup_aranged):
    # results to be filled (bus, p in kw, q in kvar)
    bus_pq = np.zeros(shape=(len(net["bus"].index), 6), dtype=float)
    b, pA, pB, pC, qA, qB, qC = np.array([]), np.array([]), np.array([]), np.array([]), \
        np.array([]), np.array([]), np.array([])

    ac = net["_options"]["ac"]
    # Todo: Voltage dependent loads
    elements = ["storage", "sgen", "load"]
    elements_3ph = ["asymmetric_load", "asymmetric_sgen"]
    for element in elements:
        sign = -1 if element in ['sgen', 'asymmetric_sgen'] else 1
        if len(net[element]):
            write_pq_results_to_element(net, net._ppc1, element, suffix="3ph")
            p_el, q_el, bus_el = get_p_q_b(net, element, suffix="3ph")
            pA = np.hstack([pA, sign * p_el/3])
            pB = np.hstack([pB, sign * p_el/3])
            pC = np.hstack([pC, sign * p_el/3])
            qA = np.hstack([qA, sign * q_el/3 if ac else np.zeros(len(p_el/3))])
            qB = np.hstack([qB, sign * q_el/3 if ac else np.zeros(len(p_el/3))])
            qC = np.hstack([qC, sign * q_el/3 if ac else np.zeros(len(p_el/3))])
            b = np.hstack([b, bus_el])
    for element in elements_3ph:
        sign = -1 if element in ['sgen', 'asymmetric_sgen'] else 1
        if len(net[element]):
            write_pq_results_to_element_3ph(net, element)
            p_el_A, q_el_A, p_el_B, q_el_B, p_el_C, q_el_C, bus_el = get_p_q_b_3ph(net, element)
            pA = np.hstack([pA, sign * p_el_A])
            pB = np.hstack([pB, sign * p_el_B])
            pC = np.hstack([pC, sign * p_el_C])
            qA = np.hstack([qA, sign * q_el_A if ac else np.zeros(len(p_el_A))])
            qB = np.hstack([qB, sign * q_el_B if ac else np.zeros(len(p_el_B))])
            qC = np.hstack([qC, sign * q_el_C if ac else np.zeros(len(p_el_C))])
            b = np.hstack([b, bus_el])

    # sum pq results from every element to be written to net['bus'] later on
    b_pp, vp_A, vq_A, vp_B, vq_B, vp_C, vq_C = _sum_by_group_nvals(b.astype(np.int64), pA, qA, pB, qB, pC, qC)
    b_ppc = bus_lookup_aranged[b_pp]
    bus_pq[b_ppc, 0] = vp_A
    bus_pq[b_ppc, 1] = vq_A
    bus_pq[b_ppc, 2] = vp_B
    bus_pq[b_ppc, 3] = vq_B
    bus_pq[b_ppc, 4] = vp_C
    bus_pq[b_ppc, 5] = vq_C
    return bus_pq


def _get_shunt_results(net, ppc, bus_lookup_aranged, bus_pq):
    ac = net["_options"]["ac"]

    b, p, q = np.array([]), np.array([]), np.array([])
    _is_elements = net["_is_elements"]
    bus_lookup = net["_pd2ppc_lookups"]["bus"]
    baseMVA = ppc["baseMVA"]
    baseZ = np.square(ppc["bus"][:, BASE_KV]) / baseMVA

    s = net["shunt"]
    if len(s) > 0:
        sidx = bus_lookup[s["bus"].values]
        shunt_is = _is_elements["shunt"]
        u_shunt = ppc["bus"][sidx, VM]
        step = s["step"]
        v_ratio = (ppc["bus"][sidx, BASE_KV] / net["shunt"]["vn_kv"].values) ** 2
        u_shunt = np.nan_to_num(u_shunt)
        use_step_table = False
        if "step_dependency_table" in s:
            if any(s.step_dependency_table):
                use_step_table = True
        if use_step_table:
            merged_df = s.merge(net.shunt_characteristic_table, left_on=['id_characteristic_table', 'step'],
                                right_on=['id_characteristic', 'step'], how='left', suffixes=('', '_char'))
            p_shunt_step = np.where(merged_df['step_dependency_table'], merged_df['p_mw_char'].values/merged_df['step'].values, merged_df['p_mw'].values).astype(np.float64)
            q_shunt_step = np.where(merged_df['step_dependency_table'], merged_df['q_mvar_char'].values/merged_df['step'].values, merged_df['q_mvar'].values).astype(np.float64)
            print(f"p_shunt_step dtype: {p_shunt_step.dtype}, q_shunt_step dtype: {q_shunt_step.dtype}")
        else:
            p_shunt_step = net["shunt"]["p_mw"].values
            q_shunt_step = net["shunt"]["q_mvar"].values
        p_shunt = u_shunt ** 2 * p_shunt_step * shunt_is * v_ratio * step
        net["res_shunt"]["p_mw"].values[:] = p_shunt
        p = np.hstack([p, p_shunt])
        if ac:
            net["res_shunt"]["vm_pu"].values[:] = u_shunt
            q_shunt = u_shunt ** 2 * q_shunt_step * shunt_is * v_ratio * step
            net["res_shunt"]["q_mvar"].values[:] = q_shunt
            q = np.hstack([q, q_shunt])
        b = np.hstack([b, s["bus"].values])

    w = net["ward"]
    if len(w) > 0:
        widx = bus_lookup[w["bus"].values]
        ward_is = _is_elements["ward"]
        u_ward = ppc["bus"][widx, VM]
        u_ward = np.nan_to_num(u_ward)
        p_ward = u_ward ** 2 * net["ward"]["pz_mw"].values * ward_is
        net["res_ward"]["p_mw"].values[:] = net["res_ward"]["p_mw"].values + p_ward
        p = np.hstack([p, p_ward])
        if ac:
            net["res_ward"]["vm_pu"].values[:] = u_ward
            q_ward = u_ward ** 2 * net["ward"]["qz_mvar"].values * ward_is
            net["res_ward"]["q_mvar"].values[:] = net["res_ward"]["q_mvar"].values + q_ward
            q = np.hstack([q, q_ward])
        b = np.hstack([b, w["bus"].values])

    xw = net["xward"]
    if len(xw) > 0:
        widx = bus_lookup[xw["bus"].values]
        xward_is = _is_elements["xward"]
        u_xward = ppc["bus"][widx, VM]
        u_xward = np.nan_to_num(u_xward)
        p_xward = u_xward ** 2 * net["xward"]["pz_mw"].values * xward_is
        net["res_xward"]["p_mw"].values[:] = net["res_xward"]["p_mw"].values + p_xward
        p = np.hstack([p, p_xward])
        if ac:
            net["res_xward"]["vm_pu"].values[:] = u_xward
            q_xward = u_xward ** 2 * net["xward"]["qz_mvar"].values * xward_is
            net["res_xward"]["q_mvar"].values[:] = net["res_xward"]["q_mvar"].values + q_xward
            q = np.hstack([q, q_xward])
        b = np.hstack([b, xw["bus"].values])

    # svc = net["svc"]  # todo: uncomment this after PandaModels net also has this key
    svc = net.get("svc", np.array([]))
    if len(svc):
        svcidx = bus_lookup[svc["bus"].values]
        svc_is = _is_elements["svc"]
        net["res_svc"].loc[svc_is, "thyristor_firing_angle_degree"] = np.rad2deg(ppc["svc"][svc_is, SVC_THYRISTOR_FIRING_ANGLE])
        p = np.hstack([p, np.zeros_like(svc["bus"].values)])
        if ac:
            net["res_svc"].loc[svc_is, "vm_pu"] = ppc["bus"][svcidx[svc_is], VM]
            net["res_svc"].loc[svc_is, "va_degree"] = ppc["bus"][svcidx[svc_is], VA]
            q_svc = ppc["svc"][:, SVC_Q]
            net["res_svc"].loc[:, "q_mvar"] = q_svc  # write all because of zeros
            net["res_svc"].loc[svc_is, "x_ohm"] = ppc["svc"][svc_is, SVC_X_PU] * baseZ[svcidx[svc_is]]
            q = np.hstack([q, q_svc])
        b = np.hstack([b, svc["bus"].values])

    # ssc = net["ssc"]  # todo: uncomment this after PandaModels net also has this key
    ssc = net.get("ssc", np.array([]))
    if len(ssc):
        sscidx = bus_lookup[ssc["bus"].values]
        ssc_is = _is_elements["ssc"]
        ssc_tb = ppc["ssc"][ssc_is, SSC_INTERNAL_BUS].real.astype(np.int64)

        net["res_ssc"].loc[ssc_is, "vm_internal_pu"] = ppc["bus"][ssc_tb, VM]
        net["res_ssc"].loc[ssc_is, "va_internal_degree"] = ppc["bus"][ssc_tb, VA]
        p = np.hstack([p, np.zeros_like(ssc["bus"].values)])
        if ac:
            net["res_ssc"].loc[ssc_is, "vm_pu"] = ppc["bus"][sscidx[ssc_is], VM]
            net["res_ssc"].loc[ssc_is, "va_degree"] = ppc["bus"][sscidx[ssc_is], VA]
            q_ssc = ppc["ssc"][:, SSC_Q]
            net["res_ssc"].loc[:, "q_mvar"] = q_ssc  # write all because of zeros
            q = np.hstack([q, q_ssc])
        b = np.hstack([b, ssc["bus"].values])

    # vsc = net["vsc"]  # todo: uncomment this after PandaModels net also has this key
    vsc = net.get("vsc", np.array([]))
    if len(vsc):
        vscidx = bus_lookup[vsc["bus"].values]
        vsc_is = _is_elements["vsc"]
        vsc_tb = ppc["vsc"][vsc_is, VSC_INTERNAL_BUS].real.astype(np.int64)
        vsc_dc_fb = ppc["vsc"][vsc_is, VSC_BUS_DC].real.astype(np.int64)
        vsc_dc_tb = ppc["vsc"][vsc_is, VSC_INTERNAL_BUS_DC].real.astype(np.int64)

        net["res_vsc"].loc[vsc_is, "vm_internal_pu"] = ppc["bus"][vsc_tb, VM]
        net["res_vsc"].loc[vsc_is, "vm_internal_dc_pu"] = ppc["bus_dc"][vsc_dc_tb, DC_VM]
        net["res_vsc"].loc[vsc_is, "vm_dc_pu"] = ppc["bus_dc"][vsc_dc_fb, DC_VM]
        p_vsc = ppc["vsc"][:, VSC_P]
        net["res_vsc"].loc[:, "p_mw"] = p_vsc  # write all because of zeros
        p = np.hstack([p, p_vsc])
        if ac:
            net["res_vsc"].loc[vsc_is, "va_internal_degree"] = ppc["bus"][vsc_tb, VA]
            net["res_vsc"].loc[vsc_is, "vm_pu"] = ppc["bus"][vscidx[vsc_is], VM]
            net["res_vsc"].loc[vsc_is, "va_degree"] = ppc["bus"][vscidx[vsc_is], VA]
            q_vsc = ppc["vsc"][:, VSC_Q]
            net["res_vsc"].loc[:, "q_mvar"] = q_vsc  # write all because of zeros
            q = np.hstack([q, q_vsc])
        b = np.hstack([b, vsc["bus"].values])

    if not ac:
        q = np.zeros(len(p))
    b_pp, vp, vq = _sum_by_group(b.astype(np.int64), p, q)
    b_ppc = bus_lookup_aranged[b_pp]

    bus_pq[b_ppc, 0] += vp
    if ac:
        bus_pq[b_ppc, 1] += vq



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from math import pi
import numpy as np
import pandas as pd
from pandapower.pypower.idx_bus import \
    BUS_I, BUS_TYPE, PD, QD, GS, BS, BUS_AREA, VM, VA, BASE_KV, ZONE, VMAX, VMIN
from pandapower.pypower.idx_gen import \
    GEN_BUS, PG, QG, QMAX, QMIN, VG, MBASE, GEN_STATUS, PMAX, PMIN
from pandapower.pypower.idx_brch import \
    F_BUS, T_BUS, BR_R, BR_X, BR_G, BR_B, RATE_A, RATE_B, RATE_C, TAP, SHIFT, BR_STATUS, ANGMIN, ANGMAX
from pandapower.pypower.idx_cost import MODEL, COST, NCOST
from pandapower.create import create_empty_network, create_buses, create_ext_grid, create_loads, \
    create_sgens, create_gens, create_lines_from_parameters, create_transformers_from_parameters, \
    create_shunts, create_ext_grid, create_pwl_costs, create_poly_costs, create_impedances
from pandapower.run import runpp

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging
logger = logging.getLogger(__name__)

ppc_elms = ["bus", "branch", "gen"]


def from_ppc(ppc, f_hz=50, validate_conversion=False, **kwargs):
    """
    This function converts pypower case files to pandapower net structure.

    INPUT:
        **ppc** (dict) -  The pypower case file.

        **f_hz** (int) - The frequency of the network, by default 50

        **validate_conversion** (bool) - If True, validate_from_ppc is run after conversion. For running the validation, the ppc must already contain the pypower powerflow results or pypower must be installed, by default False

        **kwargs** (dict) - keyword arguments for:
                            - validate_from_ppc if validate_conversion is True

                            - tap_side

                            - check_costs is passed as "check" to create_pwl_costs() and create_poly_costs()
                            
                            - branch_g_name: name of branch_g column if it exists in ppc. if not given, branch G is set to 0

    OUTPUT:
        **net** - ppc converted to pandapower net structure

    EXAMPLES:
        >>> import pandapower
        >>> from pandapower.test.converter.test_from_ppc import get_testgrids
        >>> ppc = get_testgrids('pypower_cases', 'case4gs.json')
        >>> net = pandapower.converter.from_ppc(ppc, f_hz=60)
    """
    # --- catch common failures
    if np.any(ppc['bus'][:, BASE_KV] <= 0):
        logger.info('There are false baseKV given in the pypower case file.')

    net = create_empty_network(f_hz=f_hz, sn_mva=ppc["baseMVA"])
    net._from_ppc_lookups = dict()

    _from_ppc_bus(net, ppc)
    net._from_ppc_lookups["gen"] = _from_ppc_gen(net, ppc)
    net._from_ppc_lookups["branch"] = _from_ppc_branch(net, ppc, f_hz, **kwargs)
    _from_ppc_gencost(net, ppc, net._from_ppc_lookups["gen"], check=kwargs.get("check_costs", True))

    # areas are not converted

    if validate_conversion:
        logger.setLevel(logging.DEBUG)
        if not validate_from_ppc(ppc, net, **kwargs):
            logger.error("Validation failed.")

    return net


def _from_ppc_bus(net, ppc):
    """ bus data -> create buses, sgen, load, shunt """
    # create buses
    idx_buses = create_buses(
        net, ppc['bus'].shape[0], name=ppc.get("bus_name", None),
        vn_kv=ppc['bus'][:, BASE_KV], type="b", zone=ppc['bus'][:, ZONE],
        in_service=(ppc['bus'][:, BUS_TYPE] != 4).astype(bool),
        max_vm_pu=ppc['bus'][:, VMAX], min_vm_pu=ppc['bus'][:, VMIN],
        index=ppc['bus'][:, BUS_I].astype(np.int64))

    # create loads
    is_load = (ppc['bus'][:, PD] > 0) | ((ppc['bus'][:, PD] == 0) & (ppc['bus'][:, QD] != 0))
    create_loads(net, idx_buses[is_load], p_mw=ppc['bus'][is_load, PD], q_mvar=ppc['bus'][
        is_load, QD], controllable=False)

    # create sgens
    is_sgen = ppc['bus'][:, PD] < 0
    create_sgens(net, idx_buses[is_sgen], p_mw=-ppc['bus'][is_sgen, PD], q_mvar=-ppc['bus'][
        is_sgen, QD], type="", controllable=False)

    # create shunts
    is_shunt = (ppc['bus'][:, GS] != 0) | (ppc['bus'][:, BS] != 0)
    create_shunts(net, idx_buses[is_shunt], p_mw=ppc['bus'][is_shunt, GS],
                     q_mvar=-ppc['bus'][is_shunt, BS])

    # unused data from ppc: VM, VA (partwise: in ext_grid), BUS_AREA


def _from_ppc_gen(net, ppc):
    """ gen data -> create ext_grid, gen, sgen """
    n_gen = ppc["gen"].shape[0]

    # if in ppc is only one gen -> numpy initially uses one dim array -> change to two dim array
    if len(ppc["gen"].shape) == 1:
        ppc["gen"] = np.array(ppc["gen"], ndmin=2)

    bus_pos = _get_bus_pos(ppc, ppc["gen"][:, GEN_BUS])

    # determine which gen should considered as ext_grid, gen or sgen
    is_ext_grid, is_gen, is_sgen = _gen_to_which(ppc, bus_pos=bus_pos)

    # take VG of the last gen of each bus
    vg_bus_lookup = pd.DataFrame({"vg": ppc["gen"][:, VG], "bus": bus_pos})
    # vg_bus_lookup = vg_bus_lookup.drop_duplicates(subset=["bus"], keep="last").set_index("bus")["vg"]
    vg_bus_lookup = vg_bus_lookup.drop_duplicates(subset=["bus"]).set_index("bus")["vg"]

    gen_name = ppc.get("gen_name", np.array([None]*n_gen))

    # create ext_grid
    idx_eg = list()
    for i in np.arange(n_gen, dtype=np.int64)[is_ext_grid]:
        idx_eg.append(create_ext_grid(
            net, bus=net.bus.index[bus_pos[i]], vm_pu=vg_bus_lookup.at[bus_pos[i]],
            va_degree=ppc['bus'][bus_pos[i], VA],
            in_service=(ppc['gen'][i, GEN_STATUS] > 0).astype(bool),
            max_p_mw=ppc['gen'][i, PMAX], min_p_mw=ppc['gen'][i, PMIN],
            max_q_mvar=ppc['gen'][i, QMAX], min_q_mvar=ppc['gen'][i, QMIN],
            name=gen_name[i]))

    # create gen
    idx_gen = create_gens(
        net, buses=net.bus.index[bus_pos[is_gen]], vm_pu=vg_bus_lookup.loc[bus_pos[is_gen]].values,
        p_mw=ppc['gen'][is_gen, PG], sn_mva=ppc['gen'][is_gen, MBASE],
        in_service=(ppc['gen'][is_gen, GEN_STATUS] > 0), controllable=True,
        max_p_mw=ppc['gen'][is_gen, PMAX], min_p_mw=ppc['gen'][is_gen, PMIN],
        max_q_mvar=ppc['gen'][is_gen, QMAX], min_q_mvar=ppc['gen'][is_gen, QMIN],
        name=gen_name[is_gen])

    # create sgen
    idx_sgen = create_sgens(
        net, buses=net.bus.index[bus_pos[is_sgen]], p_mw=ppc['gen'][is_sgen, PG],
        q_mvar=ppc['gen'][is_sgen, QG], sn_mva=ppc['gen'][is_sgen, MBASE], type="",
        in_service=(ppc['gen'][is_sgen, GEN_STATUS] > 0),
        max_p_mw=ppc['gen'][is_sgen, PMAX], min_p_mw=ppc['gen'][is_sgen, PMIN],
        max_q_mvar=ppc['gen'][is_sgen, QMAX], min_q_mvar=ppc['gen'][is_sgen, QMIN],
        controllable=True, name=gen_name[is_sgen])

    neg_p_gens = np.arange(n_gen, dtype=np.int64)[(ppc['gen'][:, PG] < 0) & (is_gen | is_sgen)]
    neg_p_lim_false = np.arange(n_gen, dtype=np.int64)[ppc['gen'][:, PMIN] > ppc['gen'][:, PMAX]]
    neg_q_lim_false = np.arange(n_gen, dtype=np.int64)[ppc['gen'][:, QMIN] > ppc['gen'][:, QMAX]]
    if len(neg_p_gens):
        logger.info(f'These gen have PG < 0 and are not converted to ext_grid: {neg_p_gens}.')
    if len(neg_p_lim_false):
        logger.info(f'These gen have PMIN > PMAX: {neg_p_lim_false}.')
    if len(neg_q_lim_false):
        logger.info(f'These gen have QMIN > QMAX: {neg_q_lim_false}.')

    # unused data from ppc: Vg (partwise: in ext_grid and gen), mBase, Pc1, Pc2, Qc1min, Qc1max,
    # Qc2min, Qc2max, ramp_agc, ramp_10, ramp_30,ramp_q, apf

    # gen_lookup
    gen_lookup = pd.DataFrame({"element": [-1]*n_gen, "element_type": [""]*n_gen})
    for is_, idx, et in zip([is_ext_grid, is_gen, is_sgen],
                            [idx_eg, idx_gen, idx_sgen],
                            ["ext_grid", "gen", "sgen"]):
        gen_lookup.loc[is_, "element"] = idx
        gen_lookup.loc[is_, "element_type"] = et
    return gen_lookup


def _from_ppc_branch(net, ppc, f_hz, **kwargs):
    """ branch data -> create line, trafo """
    n_bra = ppc["branch"].shape[0]

    # todo how to preserve this information (g, r_asym, x_asym, g_asym, b_asym):
    #  * for branches that are not transformers but non-zero r_asym, x_asym, g_asym, b_asym
    #       - create as impedance instead of line
    #  * for branches that are transfromers but have non-zero g_asym, b_asym: --> not done yet
    #       - write a new function to convert delta to wye
    #       - obtain the values for rft, rtf, xft, xtf
    #       - calculate ratios for HV portion of r and x
    #       - write the ratios in trafo columns leakage_resistance_ratio_hv, leakage_reactance_ratio_hv
    #  * for branches that are not transformers but connect different voltage levels:
    #       - import them as impedance instead

    zero_column = np.zeros(n_bra, dtype=np.float64)
    br_r_asym = ppc.get("branch_r_asym", zero_column)
    br_x_asym = ppc.get("branch_x_asym", zero_column)
    br_g = ppc.get("branch_g", zero_column)
    br_g_asym = ppc.get("branch_g_asym", zero_column)
    br_b_asym = ppc.get("branch_b_asym", zero_column)

    # --- general_parameters
    baseMVA = ppc['baseMVA']  # MVA
    omega = pi * f_hz  # 1/s
    MAX_VAL = 99999.

    from_bus = _get_bus_pos(ppc, ppc['branch'][:, F_BUS].astype(np.int64))
    to_bus = _get_bus_pos(ppc, ppc['branch'][:, T_BUS].astype(np.int64))
    from_vn_kv = ppc['bus'][from_bus, BASE_KV]
    to_vn_kv = ppc['bus'][to_bus, BASE_KV]

    is_line, is_trafo, is_impedance, to_vn_is_leq = _branch_to_which(ppc, from_vn_kv=from_vn_kv, to_vn_kv=to_vn_kv)

    bra_name = ppc.get("branch_name", ppc.get("bra_name", np.array([None]*n_bra)))

    # --- create line
    Zni = ppc['bus'][to_bus, BASE_KV]**2/baseMVA  # ohm
    max_i_ka = ppc['branch'][:, RATE_A]/ppc['bus'][to_bus, BASE_KV]/np.sqrt(3)
    i_is_zero = np.isclose(max_i_ka, 0)
    if np.any(i_is_zero):
        max_i_ka[i_is_zero] = MAX_VAL
        logger.debug("ppc branch rateA is zero -> Using MAX_VAL instead to calculate " +
                     "maximum branch flow")
    idx_line = create_lines_from_parameters(
        net, from_buses=net.bus.index[from_bus[is_line]], to_buses=net.bus.index[to_bus[is_line]],
        length_km=1, name=bra_name[is_line],
        r_ohm_per_km=(ppc['branch'][is_line, BR_R]*Zni[is_line]),
        x_ohm_per_km=(ppc['branch'][is_line, BR_X]*Zni[is_line]),
        c_nf_per_km=(ppc['branch'][is_line, BR_B]/Zni[is_line]/omega*1e9/2),
        g_us_per_km=(br_g[is_line]/Zni[is_line]*1e6/2),
        max_i_ka=max_i_ka[is_line], type='ol', max_loading_percent=100,
        in_service=ppc['branch'][is_line, BR_STATUS].astype(bool))

    # --- create transformer
    if np.any(is_trafo):
        hv_bus = from_bus[is_trafo]
        vn_hv_kv = from_vn_kv[is_trafo]
        lv_bus = to_bus[is_trafo]
        vn_lv_kv = to_vn_kv[is_trafo]
        if not np.all(to_vn_is_leq):
            hv_bus[~to_vn_is_leq] = to_bus[is_trafo][~to_vn_is_leq]
            vn_hv_kv[~to_vn_is_leq] = to_vn_kv[is_trafo][~to_vn_is_leq]
            lv_bus[~to_vn_is_leq] = from_bus[is_trafo][~to_vn_is_leq]
            vn_lv_kv[~to_vn_is_leq] = from_vn_kv[is_trafo][~to_vn_is_leq]
        same_vn = to_vn_kv[is_trafo] == from_vn_kv[is_trafo]
        if np.any(same_vn):
            logger.warning(
                f'There are {sum(same_vn)} branches which are considered as trafos - due to ratio '
                f'unequal 0 or 1 - but connect same voltage levels.')
        rk = ppc['branch'][is_trafo, BR_R]
        xk = ppc['branch'][is_trafo, BR_X]
        zk = (rk ** 2 + xk ** 2) ** 0.5
        sn = ppc['branch'][is_trafo, RATE_A]
        sn_is_zero = np.isclose(sn, 0)
        if np.any(sn_is_zero):
            sn[sn_is_zero] = MAX_VAL
            logger.debug("ppc branch rateA is zero -> Using MAX_VAL instead to calculate " +
                            "apparent power")
        tap_side = kwargs.get("tap_side", "hv")
        if isinstance(tap_side, str):
            tap_side_is_hv = np.array([tap_side == "hv"]*sum(is_trafo))
        else:
            tap_side_is_hv = np.array(tap_side == "hv")
        ratio_1 = ppc['branch'][is_trafo, TAP]
        ratio_is_zero = np.isclose(ratio_1, 0)
        ratio_1[~ratio_is_zero & ~tap_side_is_hv] **= -1
        ratio_1[~ratio_is_zero] -= 1
        br_b = ppc['branch'][is_trafo, BR_B]
        if np.any(br_b > 0):
            logger.warning("some BR_B of transformers in ppc['branch'] are positive, but transformers are inductive, therefore their B should be negative")
        i0_percent = np.sqrt(br_b**2 + br_g[is_trafo]**2) * 100 * baseMVA / sn
        # i0_percent = -ppc['branch'][is_trafo, BR_B] * 100 * baseMVA / sn
        is_neg_i0_percent = i0_percent < 0
        if np.any(is_neg_i0_percent):
            logger.info(
                'Transformers always behave inductive consumpting but the susceptance of pypower '
                f'branches {np.arange(len(is_neg_i0_percent), dtype=np.int64)[is_neg_i0_percent]} '
                f'(hv_bus, lv_bus)=({hv_bus[is_neg_i0_percent]}, {hv_bus[is_neg_i0_percent]}) '
                'is positive.')

        pfe_kw = br_g[is_trafo] * baseMVA * 1e3 
        # pfe_kw = 0.
        vk_percent = np.sign(xk) * zk * sn * 100 / baseMVA
        vk_percent[~tap_side_is_hv] /= (1+ratio_1[~tap_side_is_hv])**2
        vkr_percent = rk * sn * 100 / baseMVA
        vkr_percent[~tap_side_is_hv] /= (1+ratio_1[~tap_side_is_hv])**2

        tap_step_percent = np.abs(ratio_1)*100
        tap_changer_type = np.where(tap_step_percent > 0, "Ratio", None)

        idx_trafo = create_transformers_from_parameters(
            net, hv_buses=net.bus.index[hv_bus], lv_buses=net.bus.index[lv_bus], sn_mva=sn,
            vn_hv_kv=vn_hv_kv, vn_lv_kv=vn_lv_kv, name=bra_name[is_trafo],
            vk_percent=vk_percent, vkr_percent=vkr_percent,
            max_loading_percent=100, pfe_kw=pfe_kw, i0_percent=i0_percent,
            shift_degree=ppc['branch'][is_trafo, SHIFT],
            tap_step_percent=tap_step_percent, tap_pos=np.sign(ratio_1),
            tap_side=tap_side, tap_neutral=0, tap_changer_type=tap_changer_type)
    else:
        idx_trafo = []
    # unused data from ppc: rateB, rateC
    
    if np.any(is_impedance):
        fb = net.bus.index[from_bus[is_impedance]]
        tb = net.bus.index[to_bus[is_impedance]]
        sn_mva = ppc['branch'][is_impedance, RATE_A]
        sn_is_zero = np.isclose(sn_mva, 0)
        if np.any(sn_is_zero):
            sn[sn_is_zero] = MAX_VAL
            logger.debug("ppc branch rateA is zero -> Using MAX_VAL instead to calculate " +
                            "apparent power")
        # the impedances in ppc[branch] refer to the net.sn_mva and the impedances in net.impedance
        # refer to net.impedance.sn_mva!
        rft_pu = ppc['branch'][is_impedance, BR_R] / baseMVA * sn_mva
        xft_pu = ppc['branch'][is_impedance, BR_X] / baseMVA * sn_mva
        bf_pu = ppc['branch'][is_impedance, BR_B] * baseMVA / sn_mva / 2
        gf_pu = br_g[is_impedance] * baseMVA / sn_mva / 2
        r_asym_pu = br_r_asym[is_impedance] / baseMVA * sn_mva
        x_asym_pu = br_x_asym[is_impedance] / baseMVA * sn_mva
        g_asym_pu = br_g_asym[is_impedance] * baseMVA / sn_mva / 2
        b_asym_pu = br_b_asym[is_impedance] * baseMVA / sn_mva / 2

        # divide by 2 because in ppc[branch] it stands for the total Y of the branch,
        # and here we specify the "from" and "to" portions of Y separately
        idx_impedance = create_impedances(net, from_buses=fb, to_buses=tb, rft_pu=rft_pu, 
                                          xft_pu=xft_pu, rtf_pu=rft_pu+r_asym_pu, xtf_pu=xft_pu+x_asym_pu,
                                          bf_pu=bf_pu, gf_pu=gf_pu, gt_pu=gf_pu+g_asym_pu,
                                          bt_pu=bf_pu+b_asym_pu, sn_mva=sn_mva)
    else:
        idx_impedance = []
    
    # branch_lookup: which branches are lines, and which ones are transformers
    branch_lookup = pd.DataFrame({"element": [-1] * n_bra, "element_type": [""] * n_bra})
    branch_lookup.loc[is_line, "element"] = idx_line
    branch_lookup.loc[is_line, "element_type"] = "line"
    branch_lookup.loc[is_trafo, "element"] = idx_trafo
    branch_lookup.loc[is_trafo, "element_type"] = "trafo"
    branch_lookup["element"] = branch_lookup["element"].astype("float64")
    branch_lookup.loc[is_impedance, "element"] = idx_impedance
    branch_lookup.loc[is_impedance, "element_type"] = "impedance"
    return branch_lookup


def _get_bus_pos(ppc, bus_names):
    try:
        return pd.Series(np.arange(ppc["bus"].shape[0], dtype=np.int64), index=ppc["bus"][
            :, BUS_I]).loc[bus_names].values
    except:
        return pd.Series(np.arange(ppc["bus"].shape[0], dtype=np.int64), index=ppc["bus"][
            :, BUS_I].astype(np.int64)).loc[bus_names].values


def _gen_to_which(ppc, bus_pos=None, flattened=True):
    if bus_pos is None:
        bus_pos = _get_bus_pos(ppc, ppc["gen"][:, GEN_BUS])
    bus_type_df = pd.DataFrame({"bus_type": ppc["bus"][bus_pos, BUS_TYPE],
                                "bus": ppc["gen"][:, GEN_BUS]}).astype(np.int64)
    bus_type_df = pd.concat([bus_type_df.loc[bus_type_df.bus_type == 3],
                            bus_type_df.loc[bus_type_df.bus_type == 2],
                            bus_type_df.loc[bus_type_df.bus_type == 1],
                            bus_type_df.loc[bus_type_df.bus_type == 4]
                            ])
    is_ext_grid = ((bus_type_df["bus_type"] == 3) & ~bus_type_df.duplicated(subset=[
        "bus"])).sort_index().values
    is_gen = ((bus_type_df["bus_type"] == 2) & ~bus_type_df.duplicated(subset=[
        "bus"])).sort_index().values
    is_sgen = ~(is_ext_grid | is_gen) & bus_type_df["bus_type"].sort_index().isin([3, 2, 1]).values
    return is_ext_grid, is_gen, is_sgen


def _branch_to_which(ppc, from_vn_kv=None, to_vn_kv=None, flattened=True):
    if from_vn_kv is None:
        from_bus = _get_bus_pos(ppc, ppc['branch'][:, F_BUS].astype(np.int64))
        from_vn_kv = ppc['bus'][from_bus, BASE_KV]
    if to_vn_kv is None:
        to_bus = _get_bus_pos(ppc, ppc['branch'][:, T_BUS].astype(np.int64))
        to_vn_kv = ppc['bus'][to_bus, BASE_KV]
    is_line = (from_vn_kv == to_vn_kv) & \
              ((ppc['branch'][:, TAP] == 0) | (ppc['branch'][:, TAP] == 1)) & \
              (ppc['branch'][:, SHIFT] == 0)
    is_trafo = (ppc['branch'][:, TAP] != 0) & (ppc['branch'][:, TAP] != 1) | \
                (ppc['branch'][:, SHIFT] != 0)
    is_impedance = ~is_line & ~is_trafo
    to_vn_is_leq = to_vn_kv[is_trafo] <= from_vn_kv[is_trafo]
    return is_line, is_trafo, is_impedance, to_vn_is_leq


def _calc_pp_pwl_points(ppc_pwl_points):
    """
    Converts ppc pwl points which is (x1, y1, ..., xn, yn)
    into pandapower pwl points which is ((x1, x2, c12), ..., (x(n-1), xn, c(n-1)n))
    with c12 is the slope between the points x1 and x2.
    """

    def construct_list_of_list(row):
        arr = pts[row, ::2]
        arr = np.concatenate((arr[:1], np.repeat(arr[1:-1], 2), arr[-1:])).reshape((-1, 2))
        arr = np.c_[arr, c[row, :]]
        arr = arr[~np.isnan(arr[:, 2])]
        return arr.tolist()

    pts = ppc_pwl_points
    if not (pts.shape[1] % 2) == 0:
        raise ValueError("_calc_pp_pwl_points() expects ppc_pwl_points with shape[1] is "
                         f"multiple of 2. However, ppc_pwl_points.shape[1]={ppc_pwl_points}.")
    c = (pts[:, 3::2] - pts[:, 1:-2:2]) / (pts[:, 2::2] - pts[:, :-2:2])
    return [construct_list_of_list(row) for row in range(pts.shape[0])]


def _add_to_gencost(gencost, n_rows, model):
    if gencost.shape[0] == n_rows:
        return gencost
    else:
        assert n_rows > gencost.shape[0]
        to_add = np.zeros((n_rows-gencost.shape[0], gencost.shape[1]))
        to_add[:, MODEL] = model
        return np.concatenate((gencost, to_add))


def _add_pwls_to_gencost(gencost, n_rows):
    return _add_to_gencost(gencost, n_rows, 1)


def _add_polys_to_gencost(gencost, n_rows):
    return _add_to_gencost(gencost, n_rows, 2)


def _from_ppc_gencost(net, ppc, gen_lookup, check=True):
    # --- gencost -> create polynomial_cost, piecewise_cost

    if 'gencost' not in ppc:
        return

    if len(ppc['gencost'].shape) == 1:
        # reshape gencost if only one gencost is given -> no indexError
        ppc['gencost'] = ppc['gencost'].reshape((1, -1))

    # construct glu2, gc2consider, n_glu
    n_glu = gen_lookup.shape[0]
    glu2 = pd.concat([gen_lookup, gen_lookup])
    if ppc['gencost'].shape[0] > 2*n_glu:
        logger.warning(
            f"There are {ppc['gencost'].shape[0]} gencost rows. Since only {n_glu} gens are "
            f"created only the first {2*n_glu} rows in gencost are considered (for p and q).")
        gc2consider = ppc['gencost'][:2*n_glu, :]
    else:
        gc2consider = ppc['gencost']

    # check gencost type
    if not all(np.isclose(gc2consider[:, MODEL], 1) | np.isclose(gc2consider[:, MODEL], 2)):
        raise ValueError("ppc['gencost'][:, 0] must be in [1, 2].")

    # construct gc4pwl, gc4poly, is_p
    gc4pwl = _add_polys_to_gencost(gc2consider, glu2.shape[0])
    gc4poly = _add_pwls_to_gencost(gc2consider, glu2.shape[0])
    is_p = np.array([True]*n_glu + [False]*n_glu)

    # --- create pwl costs
    is_pwl = gc4pwl[:, MODEL] == 1
    for is_, power_type in zip([is_p, ~is_p], ["p", "q"]):
        is_i = is_pwl & is_
        if not sum(is_i):
            continue
        if not np.allclose(2*gc4pwl[is_i, NCOST], gc4pwl[is_i, COST:].shape[1]):
            raise ValueError("In pwl gencost, the number n does not fit to the number of values")
        pp_pwl_points = _calc_pp_pwl_points(gc4pwl[is_i, 4:])
        create_pwl_costs(net, glu2.element.values[is_i], glu2.element_type.values[is_i],
                         pp_pwl_points, power_type=power_type, check=check)

    # --- create poly costs
    is_poly = np.any(np.isclose(np.c_[gc4poly[:n_glu, MODEL], gc4poly[n_glu:, MODEL]], 2), axis=1)
    is_poly2 = np.concatenate((is_poly, is_poly))

    if sum(is_poly):
        ncost = gc4poly[:, NCOST]
        if any(is_poly2 & (ncost > 3)):
            logger.warning("The pandapower poly_cost table only supports up to 2nd order " +
                           "polynomials. The ppc higher order polynomials cannot be converted.")
            ncost[is_poly2 & (ncost > 3), NCOST] = 3
        is_cost1 = np.isclose(ncost, 1)
        is_cost2 = np.isclose(ncost, 2)
        is_cost3 = np.isclose(ncost, 3)
        if not any(is_poly2 & (~is_cost1 | ~is_cost2 | ~is_cost3)):
            raise ValueError("'ncost' must be an positve integers.")
        poly_c = {key: np.zeros((2*n_glu, )) for key in ["c0", "c1", "c2"]}
        poly_c["c0"][is_poly2 & is_cost1] = gc4poly[is_poly2 & is_cost1, COST]
        if any(is_cost2):
            poly_c["c0"][is_poly2 & is_cost2] = gc4poly[is_poly2 & is_cost2, COST+1]
            poly_c["c1"][is_poly2 & is_cost2] = gc4poly[is_poly2 & is_cost2, COST  ]
        if any(is_cost3):
            poly_c["c0"][is_poly2 & is_cost3] = gc4poly[is_poly2 & is_cost3, COST+2]
            poly_c["c1"][is_poly2 & is_cost3] = gc4poly[is_poly2 & is_cost3, COST+1]
            poly_c["c2"][is_poly2 & is_cost3] = gc4poly[is_poly2 & is_cost3, COST  ]

        create_poly_costs(
            net,
            gen_lookup.element.values[is_poly],
            gen_lookup.element_type.values[is_poly],
            cp0_eur =         poly_c["c0"][is_poly2 & is_p],
            cp1_eur_per_mw =  poly_c["c1"][is_poly2 & is_p],
            cp2_eur_per_mw2 = poly_c["c2"][is_poly2 & is_p],
            cq0_eur =           poly_c["c0"][is_poly2 & ~is_p],
            cq1_eur_per_mvar =  poly_c["c1"][is_poly2 & ~is_p],
            cq2_eur_per_mvar2 = poly_c["c2"][is_poly2 & ~is_p],
            check=check)


def _validate_diff_res(diff_res, max_diff_values):
    val = True
    for et_val in ['gen_q_mvar', 'branch_p_mw', 'branch_q_mvar', 'gen_p_mw', 'bus_va_degree',
                   'bus_vm_pu']:
        if max_diff_values[et_val] is not None:
            et = et_val.split("_")[0]
            log_key = et if et != "gen" else "gen_p" if "p" in et_val else "gen_q_sum_per_bus"
            i_col = _log_dict(log_key)[0]
            val &= bool(np.max(abs(diff_res[log_key][:, i_col])) < max_diff_values[et_val])
    return val


def _gen_q_per_bus_sum(q_array, ppc):
    return pd.DataFrame(
        np.c_[q_array, ppc["gen"][:, GEN_BUS].astype(np.int64)],
        columns=["q_mvar", "bus"]).groupby("bus").sum()


def _log_dict(key=None):
    log_dict = {
        "bus": [[0, 1], ["voltage magnitude", "voltage angle"], ["pu", "degree"]],
        "branch": [[[0, 2], [1, 3]], ["branch flow active power", "branch flow reactive power"],
                   ["MW", "Mvar"]],
        "gen_p": [[0], ["active power generation"], ["MW"]],
        "gen_q_sum_per_bus": [[0], ["reactive power generation sum per bus"], ["Mvar"]]}
    if key is None:
        return log_dict
    else:
        return log_dict[key]


def validate_from_ppc(ppc, net, max_diff_values={"bus_vm_pu": 1e-6, "bus_va_degree": 1e-5,
                                                 "branch_p_mw": 1e-6, "branch_q_mvar": 1e-6,
                                                 "gen_p_mw": 1e-6, "gen_q_mvar": 1e-6}):
    """
    This function validates the conversion of a pypower case file (ppc) to a pandapower net.
    It compares the power flow calculation results which must be provided within the ppc and the net.

    INPUT:
        **ppc** - dict

        **net** - pandapower.pandapowerNet

        **max_diff_values** - dict, optional by default { "bus_vm_pu": 1e-6, "bus_va_degree": 1e-5, "branch_p_mw": 1e-6,
            "branch_q_mvar": 1e-6, "gen_p_mw": 1e-6, "gen_q_mvar": 1e-6}

    OUTPUT:
    **pf_match** - Whether the power flow results matches.

    EXAMPLES:
        >>> import pandapower
        >>> from pandapower.test.converter.test_from_ppc import get_testgrids
        >>> ppc = get_testgrids('pypower_cases', 'case4gs.json')
        >>> net = pandapower.converter.from_ppc(ppc, f_hz=50)
        >>> pandapower.runpp(net)
        >>> pf_match = pandapower.converter.validate_from_ppc(ppc, net)
    """
    if "_from_ppc_lookups" not in net.keys() or \
            ("gen" not in net._from_ppc_lookups.keys() and len(ppc["gen"]) > 0) or \
            ("branch" not in net._from_ppc_lookups.keys() and len(ppc["branch"]) > 0):
        raise ValueError(
            "net._from_ppc_lookups must contain a lookup (dict of keys 'branch' and 'gen')")

    if net.res_bus.shape[0] == 0 and net.bus.shape[0] > 0:
        logger.debug("runpp() is performed by validate_from_ppc() since res_bus is empty.")
        runpp(net, calculate_voltage_angles=True, trafo_model="pi")

    # --- pypower powerflow results -> ppc_res -----------------------------------------------------
    ppc_res = dict.fromkeys(ppc_elms)
    ppc_res["bus"] = ppc['bus'][:, 7:9]
    ppc_res["branch"] = ppc['branch'][:, 13:17]
    ppc_res["gen"] = ppc['gen'][:, 1:3]
    ppc_res["gen_p"] = ppc_res["gen"][:, :1]
    ppc_res["gen_q_sum_per_bus"] = _gen_q_per_bus_sum(ppc_res["gen"][:, -1:], ppc)

    # --- pandapower powerflow results -> pp_res ---------------------------------------------------
    pp_res = dict.fromkeys(ppc_elms)

    # --- bus
    pp_res["bus"] = net.res_bus.loc[ppc["bus"][:, BUS_I].astype(np.int64), ['vm_pu', 'va_degree']].values

    # --- branch
    pp_res["branch"] = np.zeros(ppc_res["branch"].shape)
    from_to_buses = -np.ones((ppc_res["branch"].shape[0], 2), dtype=np.int64)
    for et in net._from_ppc_lookups["branch"].element_type.unique():
        if et == "line" or et == "impedance":
            from_to_cols = ["from_bus", "to_bus"]
            res_cols = ['p_from_mw', 'q_from_mvar', 'p_to_mw', 'q_to_mvar']
        elif et == "trafo":
            from_to_cols = ["hv_bus", "lv_bus"]
            res_cols = ['p_hv_mw', 'q_hv_mvar', 'p_lv_mw', 'q_lv_mvar']
        else:
            raise NotImplementedError(
                f"result columns for element type {et} are not implemented.")
        is_et = net._from_ppc_lookups["branch"].element_type == et
        pp_res["branch"][is_et] += net[f"res_{et}"].loc[
            net._from_ppc_lookups["branch"].element.loc[is_et], res_cols].values
        from_to_buses[is_et] = net[et].loc[
            net._from_ppc_lookups["branch"].element.loc[is_et], from_to_cols].values

    # switch direction as in ppc
    correct_from_to = np.all(from_to_buses == ppc["branch"][:, F_BUS:T_BUS+1].astype(np.int64), axis=1)
    switch_from_to = np.all(from_to_buses[:, ::-1] == ppc["branch"][:, F_BUS:T_BUS+1].astype(
        np.int64), axis=1)
    if not np.all(correct_from_to | switch_from_to):
        raise ValueError("ppc branch from and to buses don't fit to pandapower from and to + "
                        "hv and lv buses.")
    if np.any(switch_from_to):
        pp_res["branch"][switch_from_to, :] = pp_res["branch"][switch_from_to, :][:, [2, 3, 0, 1]]

    # --- gen
    pp_res["gen"] = np.zeros(ppc_res["gen"].shape)
    res_cols = ['p_mw', 'q_mvar']
    for et in net._from_ppc_lookups["gen"].element_type.unique():
        is_et = net._from_ppc_lookups["gen"].element_type == et
        pp_res["gen"][is_et] += net[f"res_{et}"].loc[
            net._from_ppc_lookups["gen"].element.loc[is_et], res_cols].values

    pp_res["gen_p"] = ppc_res["gen"][:, :1]
    pp_res["gen_q_sum_per_bus"] = _gen_q_per_bus_sum(pp_res["gen"][:, -1:], ppc)

    # --- log maximal differences the powerflow result comparison
    diff_res = dict()
    comp_keys = ["bus", "branch", "gen_p", "gen_q_sum_per_bus"]
    for comp_key in comp_keys:
        diff_res[comp_key] = ppc_res[comp_key] - pp_res[comp_key]
        if isinstance(diff_res[comp_key], pd.DataFrame):
            diff_res[comp_key] = diff_res[comp_key].values
        for i_col, var_str, unit in zip(*_log_dict(comp_key)):
            diff = diff_res[comp_key][:, i_col]
            logger.debug(f"Maximum {var_str} difference between pandapower and pypower: "
                         "%.2e %s" % (np.max(abs(diff)), unit))

    # --- do the powerflow result comparison -------------------------------------------------------
    pf_match = _validate_diff_res(diff_res, max_diff_values)

    # --- case of missmatch: result comparison with different max_diff (unwanted behaviour of
    # pypower possible)
    if not pf_match:
        other_max_diff = {
            "bus_vm_pu": 1e-3, "bus_va_degree": 1e-3, "branch_p_mw": 1e-6, "branch_q_mvar": 1e-6,
            "gen_p_mw": None, "gen_q_mvar": None}
        if _validate_diff_res(diff_res, other_max_diff) and \
                (np.max(abs(pp_res["gen"] - ppc_res["gen"])) > 1e-1).any():
            logger.debug("The active/reactive power generation difference possibly results "
                        "because of a pypower error. Please validate "
                        "the results via pypower loadflow.")  # this occurs e.g. at ppc case9

    return pf_match


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import copy
import numpy as np
import pandas as pd
import pandas.testing as pdt
import uuid

from pandapower.auxiliary import ensure_iterability, log_to_level
from pandapower.create import create_empty_network, _group_parameter_list, _set_multiple_entries, \
    _check_elements_existence, create_group
from pandapower.toolbox.power_factor import signing_system_value
from pandapower.toolbox.element_selection import branch_element_bus_dict, element_bus_tuples, \
    pp_elements, get_connected_elements_dict
from pandapower.toolbox.result_info import res_power_columns

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)

"""
Notes
-----
Using different reference_columns for the same group and element_type is not supported.
See check_unique_group_rows()
"""


# ====================================
# CREATE AND DELETE GROUPS
# ====================================

# pandapower.create.create_group


# pandapower.create.create_group_from_dict


def drop_group(net, index):
    """Drops the group of given index.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        index of the group which should be dropped
    """
    net.group = net.group.drop(index)


def drop_group_and_elements(net, index):
    """
    Drops all elements of the group and in net.group the group itself.
    """
    # functions like drop_trafos, drop_lines, drop_buses are not considered since all elements
    # should be included in elements_dict
    for et in net.group.loc[[index], "element_type"].tolist():
        idx = group_element_index(net, index, et)
        net[et].drop(idx.intersection(net[et].index), inplace=True)
        res_et = "res_" + et
        if res_et in net.keys() and net[res_et].shape[0]:
            net[res_et].drop(net[res_et].index.intersection(idx), inplace=True)
    net.group = net.group.drop(index)


# ====================================
# ADAPT GROUP MEMBERS
# ====================================


def append_to_group(*args, **kwargs):
    msg = "The name of the function append_to_group() is deprecated with pp.version >= 2.12. " + \
        "Use attach_to_group() instead."
    raise DeprecationWarning(msg)


def attach_to_groups(net, index, element_types, element_indices, reference_columns=None):
    """Appends the groups by the elements given.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : list[int]
        index of the considered group
    element_types : str or list of strings
        defines, together with 'elements', which net elements belong to the group
    element_indices : list of list of indices
        defines, together with 'element_types', which net elements belong to the group
    reference_columns : string or list of strings, optional
        If given, the elements_dict should not refer to DataFrames index but to another column.
        It is highly relevant that the reference_column exists in all DataFrames of the grouped
        elements and have the same dtype, by default None

    See Also
    --------
    attach_to_group
    """
    for idx in index:
        attach_to_group(net, idx, element_types, element_indices, reference_columns=reference_columns)


def attach_to_group(net, index, element_types, element_indices, reference_columns=None,
                    take_existing_reference_columns=True):
    """Appends the group by the elements given.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        index of the considered group
    element_types : str or list of strings
        defines, together with 'elements', which net elements belong to the group
    element_indices : list of list of indices
        defines, together with 'element_types', which net elements belong to the group
    reference_columns : string or list of strings, optional
        If given, the elements_dict should not refer to DataFrames index but to another column.
        It is highly relevant that the reference_column exists in all DataFrames of the grouped
        elements and have the same dtype, by default False
    take_existing_reference_columns : bool, optional
        Determines the behavior if the given reference_columns does not match the reference_column
        existing in net.group. If True, the existing reference_column is taken for both. If False,
        an error is raised.
    """
    if index not in net.group.index:
        raise ValueError(
            f"{index} is not in net.group.index. Correct index or used create_group() instead.")

    element_types, element_indices, reference_columns = _group_parameter_list(
        element_types, element_indices, reference_columns)

    complete_new = {col: list() for col in [
        "name", "element_type", "element_index", "reference_column"]}
    name = group_name(net, index)

    is_group_index = np.isin(net.group.index.values, index)

    for et, elm, rc in zip(element_types, element_indices, reference_columns):

        group_et = is_group_index & (net.group.element_type == et).values
        no_row = np.sum(group_et)

        # --- element entry of existing is appended
        if no_row == 1:

            # handle different reference_column
            existing_rc = net.group.reference_column.loc[group_et].at[index]
            if existing_rc != rc:
                if take_existing_reference_columns:
                    temp_gr = create_group(net, [et], [elm], reference_columns=rc)
                    set_group_reference_column(net, temp_gr, existing_rc, element_type=et)
                    elm = net.group.element_index.at[temp_gr]
                    net.group = net.group.drop(temp_gr)
                else:
                    raise UserWarning(
                        f"The reference column of existing group {index} for element "
                        f"type '{et}' and of the elements to append differ. Use "
                        "set_group_reference_column() to change the reference column of net.group "
                        "before, or pass appropriate data to attach_to_group().")

            # append
            prev_elm = net.group.element_index.loc[group_et].at[index]
            prev_elm = [prev_elm] if isinstance(prev_elm, str) or not hasattr(
                prev_elm, "__iter__") else list(prev_elm)
            net.group.iat[np.arange(len(group_et), dtype=int)[group_et][0],
                          net.group.columns.get_loc("element_index")] = \
                prev_elm + list(pd.Index(elm).difference(pd.Index(prev_elm)))

        # --- prepare adding new rows to net.group (because no other elements of element type et
        # --- already belong to the group)
        elif no_row == 0:
            complete_new["name"].append(name)
            complete_new["element_type"].append(et)
            complete_new["element_index"].append(elm)
            complete_new["reference_column"].append(rc)

        else:
            raise ValueError(f"Multiple {et} rows for group {index}")

    # --- add new rows to net.group
    if len(complete_new["name"]):
        _check_elements_existence(net, element_types, element_indices, reference_columns)
        _set_multiple_entries(net, "group", [index]*len(complete_new["name"]), **complete_new)
        net.group.sort_index(inplace=True)


def drop_from_group(*args, **kwargs):
    msg = ("The name of the function drop_from_group() is deprecated with pp.version >= 2.12. "
           "Use detach_from_group() instead.")
    raise DeprecationWarning(msg)


def detach_from_group(net, index, element_type, element_index):
    """Detaches elements from the group with the given group index 'index'.
    No errors are raised if elements are passed to be drop from groups which alread don't have these
    elements as members.
    A reverse function is available -> pp.group.attach_to_group().

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        Index of the group from which the element should be dropped
    element_type : str
        The element type of which elements should be dropped from the group(s), e.g. "bus"
    element_index : int or list of integers
        indices of the elements which should be dropped from the group
    """
    detach_from_groups(net, element_type, element_index, index=index)


def drop_from_groups(*args, **kwargs):
    msg = ("The name of the function drop_from_groups() is deprecated with pp.version >= 2.12. "
           "Use detach_from_groups() instead.")
    raise DeprecationWarning(msg)


def detach_from_groups(net, element_type, element_index, index=None):
    """Detaches elements from one or multiple groups, defined by 'index'.
    No errors are raised if elements are passed to be dropped from groups which alread don't have
    these elements as members.
    A reverse function is available -> pp.group.attach_to_group().

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    element_type : str
        The element type of which elements should be dropped from the group(s), e.g. "bus"
    element_index : int or list of integers
        indices of the elements which should be dropped from the group
    index : int or list of integers, optional
        Indices of the group(s) from which the element should be dropped. If None, the elements are
        dropped from all groups, by default None
    """
    if index is None:
        index = net.group.index
    element_index = pd.Index(ensure_iterability(element_index), dtype=np.int64)

    to_check = np.isin(net.group.index.values, index)
    to_check &= net.group.element_type.values == element_type
    keep = np.ones(net.group.shape[0], dtype=bool)

    for i in np.arange(len(to_check), dtype=np.int64)[to_check]:
        rc = net.group.reference_column.iat[i]
        if rc is None or pd.isnull(rc):
            net.group.element_index.iat[i] = pd.Index(net.group.element_index.iat[i]).difference(
                element_index).tolist()
        else:
            net.group.element_index.iat[i] = pd.Index(net.group.element_index.iat[i]).difference(
                pd.Index(net[element_type][rc].loc[element_index.intersection(
                    net[element_type].index)])).tolist()

        if not len(net.group.element_index.iat[i]):
            keep[i] = False
    net.group = net.group.loc[keep]


# =================================================
# ACCESS GROUP DATA AND EVALUATE MEMBERSHIP
# =================================================


def _get_lists_from_df(df, cols):
    return [df[col].tolist() for col in cols]


def group_element_lists(net, index):
    return tuple(_get_lists_from_df(net.group.loc[[index]],
                                    ["element_type", "element_index", "reference_column"]))


def group_name(net, index):
    """Returns the name of the group and checks that all group rows include the same name

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        index of the group
    verbose : bool, optional
        Setting to False, accelerate the code but don't check inconsistent names, by default True

    See also
    --------
    group_index : recursive function
    """
    names = net.group.name.loc[[index]]
    if len(set(names)) != 1 and not pd.isnull(names).all():
        raise ValueError(f"group {index} has different values in net.group.name.loc[index]")
    return names.values[0]


def group_index(net, name):
    """Returns the index of the group named as given

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    name : str
        name of the requested group

    See also
    --------
    group_name : recursive function
    """
    index = net.group.index[net.group.name == name]
    if len(set(index)) != 1:
        raise ValueError(f"group {name} has multiple indices: {index}")
    return index[0]


def group_element_index(net, index, element_type):
    """Returns the indices of the elements of the group in the element table net[element_type]. This
    function considers net.group.reference_column.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        Index of the group
    element_type : str
        name of the element table to which the returned indices of the elements of the group belong
        to

    Returns
    -------
    pd.Index
        indices of the elements of the group in the element table net[element_type]
    """
    if element_type not in net.group.loc[[index], "element_type"].values:
        return pd.Index([], dtype=np.int64)

    row = group_row(net, index, element_type)
    element = row.at["element_index"]
    reference_column = row.at["reference_column"]

    if reference_column is None or pd.isnull(reference_column):
        return pd.Index(element, dtype=np.int64)

    return net[element_type].index[net[element_type][reference_column].isin(element)]


def group_row(net, index, element_type):
    """Returns the row which consists the data of the requested group index and element type.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        index of the group
    element_type : str
        element type (defines which row of the data of the group should be returned)

    Returns
    -------
    pandas.Series
        data of net.group, defined by the index of the group and the element type

    Raises
    ------
    KeyError
        Now row exist for the requested group and element type
    ValueError
        Multiple rows exist for the requested group and element type
    """
    group_df = net.group.loc[[index]].set_index("element_type")
    try:
        row = group_df.loc[element_type]
    except KeyError:
        raise KeyError(f"Group {index} has no {element_type}s.")
    if isinstance(row, pd.Series):
        return row
    elif isinstance(row, pd.DataFrame):
        raise ValueError(f"Multiple {element_type} rows for group {index}")
    else:
        raise ValueError(f"Returning row {element_type} for group {index} failed.")


def _single_element_index(element_index):
    if not hasattr(element_index, "__iter__"):
        return [element_index], True
    else:
        return list(element_index), False


def isin_group(net, element_type, element_index, index=None, drop_empty_lines=True):
    """Returns whether elements are in group(s).

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    element_type : str
        element type of the elements to be found in the groups, e.g. "gen" or "load"
    element_index : int or list of integers
        indices of the element table which should be found in the groups
    index : int or list of integers, optional
        Can narrow the number of groups in which the elements are searched, by default None
    drop_empty_lines : bool, optional
        This parameter decides whether empty entries should be removed (the complete row in
        net.group), by default True

    Returns
    -------
    boolean or boolean numpy.array
        Information whether the element are in any group
    """
    if isinstance(element_index, str):
        raise ValueError("element_index must be an integer or a list of integers.")
    element_index, single_element = _single_element_index(element_index)
    if index is None:
        index = list(set(net.group.index))
    else:
        index = ensure_iterability(index)

    ensure_lists_in_group_element_column(net, drop_empty_lines=drop_empty_lines)

    member_idx = pd.Index([], dtype=np.int64)
    for idx in index:
        member_idx = member_idx.union(group_element_index(net, idx, element_type))

    isin = np.isin(element_index, member_idx)

    if single_element:
        return isin[0]
    else:
        return isin


def element_associated_groups(net, element_type, element_index, return_empties=True,
                              drop_empty_lines=True):
    """Returns to which groups given elements belong to.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    element_type : str
        element type of the elements to be found in the groups, e.g. "gen" or "load"
    element_index : str
        indices of the element table which should be found in the groups
    return_empties : bool, optional
        whether entries with an empty list of assiciated groups should be returned, by default True
    drop_empty_lines : bool, optional
        This parameter decides whether empty entries should be removed (the complete row in
        net.group), by default True

    Returns
    -------
    dict[int, list[int]]
        for each element index a list of associated group indices is returned as a dict
    """
    element_index, single_element = _single_element_index(element_index)
    index = list(set(net.group.index))
    ensure_lists_in_group_element_column(net, drop_empty_lines=drop_empty_lines)
    gr_et = net.group.loc[net.group.element_type == element_type]
    associated = pd.Series(dict.fromkeys(element_index, list()))
    for idx in gr_et.index:
        ass = pd.Index(element_index).intersection(group_element_index(net, idx, element_type))
        associated.loc[ass] = associated.loc[ass].apply(lambda x: x + [idx])
    if not return_empties:
        associated = associated.loc[associated.apply(len).astype(bool)]
    associated = associated.to_dict()
    if single_element:
        return associated[index[0]]
    else:
        return associated


def count_group_elements(net, index):
    """Returns a Series concluding the number of included elements in self.elements_dict

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        index of the considered group

    Returns
    -------
    pd.Series
        number of elements per element type existing in the group

    See also
    --------
    count_elements
    """
    return pd.Series({
        et: len(elm) if hasattr(elm, "__iter__") and not isinstance(elm, str) else 1 for
        et, elm in zip(*_get_lists_from_df(
            net.group.loc[[index]], ["element_type", "element_index"]))}, dtype=np.int64)


# =================================================
# COMPARE GROUPS
# =================================================


def groups_equal(net, index1, index2, **kwargs):
    """Returns a boolean whether both group are equal.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index1 : int
        index of the first group to compare
    index2 : int
        index of the second group to compare
    """
    df1 = net.group.loc[[index1]].set_index("name")
    df2 = net.group.loc[[index2]].set_index("name")
    try:
        pdt.assert_frame_equal(df1, df2, **kwargs)
        return True
    except AssertionError:
        return False


def compare_group_elements(net, index1, index2):
    """allow_cross_reference_column_comparison

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index1 : int
        index of the first group to compare
    index2 : int
        index of the second group to compare
    """
    ensure_lists_in_group_element_column(net)
    et1 = net.group.loc[[index1], "element_type"].tolist()
    et2 = net.group.loc[[index2], "element_type"].tolist()
    if len(et1) != len(set(et1)):
        raise ValueError(f"The Group of index {index1} has duplicated element_types.")
    if len(et2) != len(set(et2)):
        raise ValueError(f"The Group of index {index2} has duplicated element_types.")
    if set(et1) != set(et2):
        return False

    gr1 = net.group.loc[[index1]].set_index("element_type")
    gr2 = net.group.loc[[index2]].set_index("element_type")
    for et in et1:
        if gr1.reference_column.at[et] == gr2.reference_column.at[et]:
            if len(pd.Index(gr1.element_index.at[et]).symmetric_difference(gr2.element_index.at[et])):
                return False
        else:
            if len(group_element_index(net, index1, et).symmetric_difference(group_element_index(
                    net, index2, et))):
                return False
    return True

# =================================================
# FIX GROUP DATA
# =================================================


def check_unique_group_names(*args, **kwargs):
    msg = ("Function check_unique_group_names() is deprecated with pp.version >= 2.12. "
           "It is replaced by check_unique_group_rows() and the raise_ parameter defaults to True.")
    raise DeprecationWarning(msg)


def check_unique_group_rows(net, raise_error=True, log_level="warning"):
    """Checks whether all groups have unique names. raise_error decides whether duplicated names lead
    to error or log message.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    raise_error : bool, optional
        decides whether duplicated names lead to error or log message., by default False
    log_level : str, optional
        the level for logs, relevant if raise_error is False

    Notes
    -----
    Using different reference_columns for the same group and element_type is not supported.
    """
    df = net.group[["name", "element_type"]].reset_index()
    if df.duplicated().any():
        raise ValueError("There are multiple group rows with same index, name and element_type.")
    single_name_per_index = [len(names) == 1 for names in net.group.reset_index().groupby("index")[
        "name"].agg(set)]
    if not all(single_name_per_index):
        warn = "Groups with different names have the same index."
        if raise_error:
            raise UserWarning(warn)
        else:
            log_to_level(warn, logger, log_level)


def remove_not_existing_group_members(net, verbose=True):
    """Remove group members from net.group that do not exist in the elements tables.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    verbose : bool, optional
        Steers amount of logging messages, by default True
    """
    gr_idx_before = set(net.group.index)
    keep = np.ones(net.group.shape[0], dtype=bool)
    not_existing_et = set()
    empty_et = set()
    for i in range(net.group.shape[0]):
        et = net.group.element_type.iat[i]
        if et not in net.keys() or not isinstance(net[et], pd.DataFrame):
            not_existing_et |= {et}
            keep[i] = False
        elif not net[et].shape[0]:
            empty_et |= {et}
            keep[i] = False
        else:
            not_exist_bool = ~group_entries_exist_in_element_table(
                net, net.group.index[i], et)
            if np.all(not_exist_bool):
                keep[i] = False
                if verbose:
                    logger.info(f"net.group row {i} is dropped because no fitting elements exist in"
                                f" net[{et}].")
            elif np.any(not_exist_bool):
                net.group.element_index.iat[i] = list(np.array(net.group.element_index.iat[i])[
                    ~not_exist_bool])
                if verbose:
                    logger.info(
                        f"{np.sum(not_exist_bool)} entries were dropped from net.group row {i}.")
    if verbose:
        if len(not_existing_et):
            logger.info(f"element_types {not_existing_et} are no dataframes in net and thus be "
                        "removed from net.group.")
        if len(empty_et):
            logger.info(f"net[*] are empty and thus be removed from net.group. "
                        f"* is placeholder for {empty_et}.")
    net.group = net.group.loc[keep]

    if verbose:
        gr_idx_after = set(net.group.index)
        removed_gr = gr_idx_before - gr_idx_after
        if len(removed_gr):
            logger.info(f"These groups are removed since no existing member remains: {removed_gr}.")


def ensure_lists_in_group_element_column(net, drop_empty_lines=True):
    """Ensure that all entries in net.group.element are of type list.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    drop_empty_lines : bool, optional
        This parameter decides whether empty entries should be removed (the complete row in
        net.group), by default True
    """
    keep = np.ones(net.group.shape[0], dtype=bool)
    for i in range(net.group.shape[0]):
        elm = net.group.element_index.iat[i]
        if hasattr(elm, "__iter__") and not isinstance(elm, str):
            net.group.element_index.iat[i] = list(elm)
            if not len(elm):
                keep[i] = False
        else:
            if elm is None or pd.isnull(elm):
                net.group.element_index.iat[i] = []
                keep[i] = False
            else:
                net.group.element_index.iat[i] = [elm]
    if drop_empty_lines:
        net.group = net.group.loc[keep]


def group_entries_exist_in_element_table(net, index, element_type):
    """Returns an array of booleans whether the entries in net.group.element exist in
    net[element_type], also considering reference_column

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        Index of the group
    element_type : str
        element type which entries should be checked, e.g. "bus"

    Returns
    -------
    array of booleans
        Whether the entries in net.group.element exist in net[element_type]
    """
    row = group_row(net, index, element_type)
    element = row.at["element_index"]
    reference_column = row.at["reference_column"]

    if not hasattr(element, "__iter__") or isinstance(element, str):
        raise ValueError("Entries in net.group.element should be lists. You can try "
                         "ensure_lists_in_group_element_column() to fix this.")

    if reference_column is None or pd.isnull(reference_column):
        return np.isin(np.array(element), net[element_type].index.values)
    else:
        return np.isin(np.array(element), net[element_type][reference_column].values)


# =================================================
# FURTHER GROUP FUNCTIONS
# =================================================


def set_group_in_service(net, index):
    """Sets all elements of the group in service.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        index of the considered group
    """
    set_value_to_group(net, index, True, "in_service", replace=True, append_column=False)


def set_group_out_of_service(net, index):
    """Sets all elements of the group out of service.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        index of the considered group
    """
    set_value_to_group(net, index, False, "in_service", replace=True, append_column=False)


def set_value_to_group(net, index, value, column, replace=True, append_column=True):
    """Sets the same value to the column of the element tables of all elements/members of the group.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        index of the considered group
    value : int/float/object
        value to be written to all group members into the element tables
    column : str
        column to be manipulated
    replace : bool, optional
        If False, value is only written to places where no value exist before (column doesn't exist
        before or value is nan), by default True
    append_column : bool, optional
        Decides whether the column should be added to element tables where this doesn't exist
        before, by default True
    """
    for et, elm, rc in zip(*group_element_lists(net, index)):
        if append_column or column in net[et].columns:
            if rc is None or pd.isnull(rc):
                if replace or column not in net[et].columns:
                    ix = elm
                else:
                    ix = net[et].loc[elm].index[net[et][column].loc[elm].isnull()]
            else:
                if replace or column not in net[et].columns:
                    ix = net[et].index[net[et][rc].isin(elm)]
                else:
                    ix = net[et].index[net[et][rc].isin(elm) & net[et][column].isnull()]
            net[et].loc[ix, column] = value


def _sum_powers(net, index, formula_character, unit):
    power = 0.
    missing_res_idx = list()
    no_power_column_found = list()
    no_res_table_found = list()
    for et in net.group.loc[[index], "element_type"].tolist():
        if et in ["switch", "measurement", "bus", "bus_geodata", "line_geodata"]:
            continue
        idx = group_element_index(net, index, et)
        res_et = "res_" + et
        if res_et not in net.keys():
            no_res_table_found.append(et)
            continue
        res_idx = net[res_et].index.intersection(idx)
        sign = 1 if et not in ["ext_grid", "gen", "sgen"] else -1
        if len(res_idx) != len(idx):
            missing_res_idx.append(et)
        col1 = "%s_%s" % (formula_character, unit)
        col2 = "%sl_%s" % (formula_character, unit)
        if col1 in net[res_et].columns:
            power += sign * net[res_et][col1].loc[res_idx].sum()
        elif col2 in net[res_et].columns:
            power += sign * net[res_et][col2].loc[res_idx].sum()
        else:
            no_power_column_found.append(et)

    if len(no_res_table_found):
        logger.warning(f"Result tables of this elements does not exist: {no_res_table_found}.")
    if len(missing_res_idx):
        logger.warning("The resulting power may be wrong since the result tables of these "
                        f"elements lack of indices: {missing_res_idx}.")
    if len(no_power_column_found):
        logger.warning("The resulting power may be wrong since the result tables of these "
                        f"elements have no requested power column: {no_power_column_found}.")
    return power


def group_res_power_per_bus(net, index):
    """Returns active and reactive power consumptions of given group per bus

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        _description_

    Returns
    -------
    pd.DataFrame
        power consumption of the group per bus (index=bus, columns=["p_mw", "q_mvar"])

    Examples
    --------
    >>> from pandapower import runpp, create_group, group_res_power_per_bus
    >>> from pandapower.networks import create_cigre_network_mv
    >>> net = create_cigre_network_mv(with_der="all")
    >>> runpp(net)
    >>> gr_idx = create_group(net, ["sgen", "line"], [[0, 1], [0, 1]], name="test group")
    >>> group_res_power_per_bus(net, gr_idx)
                 p_mw        q_mvar
    bus
    1    2.953004e+00  1.328978e+00
    2   -3.641532e-14  4.618528e-14
    3   -2.875066e+00 -1.318864e+00
    4   -2.000000e-02  0.000000e+00
    """
    pq_sums = pd.DataFrame({'p_mw': float(), 'q_mvar': float()}, index=[])
    bra_ets = pp_elements(bus=False, bus_elements=False, other_elements=False)
    bra_ebd = branch_element_bus_dict()
    missing_res_idx = list()

    for et in net.group.loc[[index], "element_type"].tolist():
        if et in ["switch", "measurement", "bus", "bus_geodata", "line_geodata"]:
            continue
        gr_elm_idx = group_element_index(net, index, et)
        sides = [0] if et not in bra_ets else [0, 1] if et != "trafo3w" else [0, 1, 2]
        for side in sides:
            bus_col = "bus" if et not in bra_ets else bra_ebd[et][side]
            idx = net[f'res_{et}'].index.intersection(gr_elm_idx)
            if len(idx) != len(gr_elm_idx):
                missing_res_idx.append(et)
            cols = res_power_columns(et, side=side)
            pq_sum = pd.concat([net[et][bus_col].loc[idx], net[f'res_{et}'][cols].loc[idx]],
                               axis=1).groupby(bus_col).sum() * signing_system_value(et)
            pq_sum.index.name = "bus"  # needs to be set for branch elements
            if len(pq_sums.columns.difference(pq_sum.columns)):
                cols_repl = {old: "p_mw" if "p" in old else "q_mvar" for old in pq_sum.columns}
                pq_sum = pq_sum.rename(columns=cols_repl)
            pq_sums = pq_sums.add(pq_sum, fill_value=0)

    if len(missing_res_idx):
        logger.warning("The resulting power may be wrong since the result tables of these "
                        f"elements lack of indices: {missing_res_idx}.")
    return pq_sums


def group_res_p_mw(net, index):
    """Sums all result table values `p_mw` of group members.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        index of the considered group

    Examples
    --------
    >>> from pandapower import runpp, create_group, group_res_p_mw
    >>> from pandapower.networks import create_cigre_network_mv
    >>> net = create_cigre_network_mv(with_der="all")
    >>> runpp(net)
    >>> gr_idx = create_group(net, ["sgen", "line"], [[0, 1], [0, 1]], name="test group")
    >>> net.res_line.pl_mw.loc[[0, 1]].sum() - net.res_sgen.p_mw.loc[[0, 1]].sum()  # expected value
    0.057938
    >>> group_res_p_mw(net, gr_idx)
    0.057938
    """
    return _sum_powers(net, index, "p", "mw")


def group_res_q_mvar(net, index):
    """Sums all result table values `q_mvar` of group members.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        index of the considered group

    Examples
    --------
    >>> from pandapower import runpp, create_group, group_res_q_mvar
    >>> from pandapower.networks import create_cigre_network_mv
    >>> net = create_cigre_network_mv(with_der="all")
    >>> runpp(net)
    >>> gr_idx = create_group(net, ["sgen", "line"], [[0, 1], [0, 1]], name="test group")
    >>> net.res_line.ql_mvar.loc[[0, 1]].sum() - net.res_sgen.q_mvar.loc[[0, 1]].sum()  # expected value
    0.010114
    >>> group_res_q_mvar(net, gr_idx)
    0.010114
    """
    return _sum_powers(net, index, "q", "mvar")


def set_group_reference_column(net, index, reference_column, element_type=None):
    """Set a reference_column to the group of given index. The values in net.group.element get
    updated.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        Index of the group
    reference_column : str
        column in the elemt tables which should be used as reference to link the group members.
        If no column but the index should be used as the link (is the default), reference_column
        should be None.
    element_type : str, optional
        Type of element which should get a new column to reference. If None, all element types are
        considered, by default None

    Raises
    ------
    ValueError
        net[element_type][reference_column] has duplicated values.
    """
    if element_type is None:
        element_type = net.group.loc[[index], "element_type"].tolist()
    else:
        element_type = ensure_iterability(element_type)

    dupl_elements = list()
    for et in element_type:

        if reference_column is None:
            # determine duplicated indices which would corrupt Groups functionality
            if len(set(net[et].index)) != net[et].shape[0]:
                dupl_elements.append(et)

        else:
            # fill nan values in net[et][reference_column] with unique names
            if reference_column not in net[et].columns:
                net[et][reference_column] = pd.Series([None]*net[et].shape[0], dtype=object)
            if pd.api.types.is_object_dtype(net[et][reference_column]):
                idxs = net[et].index[net[et][reference_column].isnull()]
                net[et].loc[idxs, reference_column] = ["%s_%i_%s" % (et, idx, str(
                    uuid.uuid4())) for idx in idxs]

            # determine duplicated values which would corrupt Groups functionality
            if (net[et][reference_column].duplicated() | net[et][reference_column].isnull()).any():
                dupl_elements.append(et)

        # update net.group[["element_index", "reference_column"]] for element_type == et
        if not len(dupl_elements):
            pos_bool = ((net.group.index == index) & (net.group.element_type == et)).values
            if np.sum(pos_bool) > 1:
                raise ValueError(
                    f"Group of index {index} has multiple entries for element type '{et}'.")
            pos = np.arange(len(pos_bool), dtype=np.int64)[pos_bool][0]

            if reference_column is None:
                net.group.iat[pos, net.group.columns.get_loc("element_index")] = \
                    group_element_index(net, index, et).tolist()
                net.group.iat[pos, net.group.columns.get_loc("reference_column")] = None
            else:
                net.group.iat[pos, net.group.columns.get_loc("element_index")] = \
                    net[et].loc[group_element_index(net, index, et), reference_column].tolist()
                net.group.iat[pos, net.group.columns.get_loc("reference_column")] = reference_column

    if len(dupl_elements):
        if reference_column is None:
            raise ValueError(f"In net[*].index have duplicated or nan values. "
                            f"* is placeholder for {dupl_elements}.")
        else:
            raise ValueError(f"In net[*].{reference_column} have duplicated or nan values. "
                            f"* is placeholder for {dupl_elements}.")


def return_group_as_net(net, index, keep_everything_else=False, verbose=True, **kwargs):
    """Returns a pandapower net consisting of the members of this group.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    index : int
        index of the considered group
    keep_everything_else : bool, optional
        Decides whether other data than element tables are kept, by default False
    verbose : bool, optional
        Decides whether logging messages are triggered, by default True

    Returns
    -------
    pandapowerNet
        pandapower net, only with elements that are members of the group
    """
    if keep_everything_else:
        group_net = copy.deepcopy(net)
        for et in pp_elements():
            if group_net[et].shape[0]:
                keep = group_element_index(net, index, et)
                group_net[et].drop(group_net[et].index.difference(keep), inplace=True)
        if len(set(net.group.index)) > 1:
            if verbose:
                logger.warning("The returned net includes further groups. Not existing members get "
                               "dropped now.")
            remove_not_existing_group_members(net, verbose=verbose)
    else:
        group_net = create_empty_network(
            name=group_name(net, index), f_hz=net.f_hz, sn_mva=net.sn_mva,
            add_stdtypes=kwargs.get("add_stdtypes", True))
        group_net["group"] = net.group.loc[[index]]
        for et in net.group.loc[[index], "element_type"].tolist():
            idx = group_element_index(net, index, et)
            group_net[et] = net[et].loc[idx]
    return group_net


def elements_connected_to_group(net, index, element_types, find_buses_only_from_buses=False,
                                respect_switches=True, respect_in_service=False,
                                include_empty_lists=False):
    """Returns a dict of indices of elements that are connected to the group.

    Parameters
    ----------
    net : pandapowerNet
        the net of the group
    index : int
        index of the group
    element_types : iterable of element types
        element types of which connected elements are searched
    find_buses_only_from_buses : bool, optional
        if True, connected buses are not searched considering bus elements and branch elements of
        the group but only considering the buses of the group as done by
        pandapower.toolbox.get_connected_buses(). In that case it is ignored whether branch
        elements between the buses of the groups and the connected buses are in the group or not
    respect_switches : bool, optional
        True -> open switches will be respected,
        False -> open switches will be ignored,
        by default True
    respect_in_service : bool, optional
        True -> in_service status of connected lines will be respected,
        False -> in_service status will be ignored,
        by default False
    include_empty_lists : bool, optional
        if True, the output doesn't have values of empty lists but may lack of element types as
        keys, by default False

    Returns
    -------
    dict[str, pd.Index]
        elements that are connected to the group
    """
    def element_type_for_switch_et(element_type):
        return element_type[0] if element_type != "trafo3w" else "t3"

    # bus->bus (for find_buses_only_from_buses) and bus->other elements connections
    group_buses = group_element_index(net, index, "bus")
    if respect_in_service:
        group_buses = net.bus.loc[group_buses].index[net.bus.in_service.loc[group_buses]]
    connected = get_connected_elements_dict(
        net, group_buses, element_types=element_types, respect_switches=respect_switches,
        respect_in_service=respect_in_service, include_empty_lists=include_empty_lists)

    # switch -> branch connections
    group_sw = group_element_index(net, index, "bus")
    sw_bra_types = ["line", "trafo", "trafo3w"]
    for et in sw_bra_types:
        if et not in element_types:
            continue
        elms = net.switch.loc[group_sw, "element"].loc[
            net.switch.loc[group_sw, "et"] == element_type_for_switch_et(et)]
        if respect_in_service:
            elms = net[et].loc[elms].index[net[et].in_service.loc[elms]]
        connected[et] = set(connected[et]) | set(elms)

    # branch -> switch connections
    if "switch" in element_types:
        conn_sw = set(connected.get("switch", set()))
        for branch_type in sw_bra_types:
            if net[branch_type].shape[0]:
                group_branches = group_element_index(net, index, branch_type)
                if respect_in_service:
                    group_branches = group_branches.intersection(
                        net[branch_type].index[net[branch_type].in_service])
                conn_sw |= set(net.switch.index[
                    (net.switch.et == element_type_for_switch_et(branch_type)) &
                    net.switch.element.isin(group_branches)])
        connected["switch"] = conn_sw

    if not find_buses_only_from_buses:
        bed = branch_element_bus_dict(include_switch=True)
        bed["switch"].append("element")
        branch_types = list(bed.keys())
        conn_buses = set()
        bed.update({tpl[0]: [tpl[1]] for tpl in element_bus_tuples(branch_elements=False)})

        for row in net.group.loc[index].itertuples():
            et = row.element_type
            if et == "bus":
                continue
            for bus_col in bed[et]:
                if et == "switch" and bus_col == "element":
                    bed_buses = net[et][bus_col].loc[net.switch.index[
                        net.switch.et == "b"].intersection(row.element_index)]
                else:
                    bed_buses = net[et][bus_col].loc[row.element_index]
                if respect_in_service and "in_service" in net[et].columns:
                    bed_buses = bed_buses.loc[net[et].in_service.loc[row.element_index]]
                if respect_switches:
                    if et == "switch":
                        bed_buses = bed_buses.loc[net.switch.closed.loc[bed_buses.index]]
                    elif et in branch_types:
                        closed = np.ones(bed_buses.shape[0], dtype=bool)
                        switches = net.switch[["bus", "closed"]].loc[
                            (net.switch.et == element_type_for_switch_et(et)) &
                            net.switch.element.isin(row.element_index)]
                        if switches.shape[0]:
                            if switches.bus.duplicated().any():
                                raise ValueError(
                                    f"There are multiple {et} switches connecting the same "
                                    "element and bus. respect_switches is not possible due to "
                                    "multiple possible values.")
                            switches = switches.set_index("bus").closed
                            in_sw = bed_buses.isin(switches.index).values
                            closed[in_sw] = switches.loc[bed_buses.loc[in_sw]]
                            bed_buses = bed_buses.loc[closed]
                conn_buses |= set(bed_buses.values)
        if respect_in_service:
            conn_buses = pd.Index(conn_buses)
            conn_buses = conn_buses[net.bus.in_service.loc[conn_buses].values]
        connected["bus"] = conn_buses

    connected = {et: sorted(pd.Index(conn).difference(group_element_index(net, index, et))) for et,
                 conn in connected.items()}
    if include_empty_lists:
        return connected
    else:
        return {key: val for key, val in connected.items() if len(val)}


if __name__ == "__main__":
    from pandapower import create_buses, create_gens, create_group, count_group_elements

    net = create_empty_network()
    create_buses(net, 3, 10)
    create_gens(net, [0]*5, [10]*5)
    create_group(net, ["bus", "gen"], [[2, 1], [1, 2]], name="hello")
    create_group(net, "bus", [[0]], name="hello")
    print(net.group)
    print(count_group_elements(net, 0))



# -*- coding: utf-8 -*-

"""
Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
and Energy System Technology (IEE), Kassel. All rights reserved.

"""

import math
import numpy as np
from itertools import product

from pandapower.auxiliary import _select_is_elements_numba, _sum_by_group
from pandapower.pypower.idx_bus_dc import DC_NONE, DC_BUS_TYPE
from pandapower.build_bus import _build_bus_ppc, _build_svc_ppc, _build_ssc_ppc, _build_vsc_ppc, _build_bus_dc_ppc
from pandapower.build_gen import _build_gen_ppc
from pandapower.pypower.idx_brch import BR_G, BR_B, BR_R, BR_X, F_BUS, T_BUS, branch_cols, BR_STATUS, SHIFT, TAP, BR_R_ASYM, \
    BR_X_ASYM, BR_G_ASYM, BR_B_ASYM
from pandapower.pypower.idx_bus import BASE_KV, BS, GS, BUS_TYPE, NONE
from pandapower.pypower.idx_brch_sc import branch_cols_sc
from pandapower.pypower.idx_bus_sc import C_MAX, C_MIN
from pandapower.build_branch import _calc_tap_from_dataframe, _transformer_correction_factor, \
    _calc_nominal_ratio_from_dataframe, \
    get_trafo_values, _trafo_df_from_trafo3w, _calc_branch_values_from_trafo_df, _calc_switch_parameter, \
    _calc_impedance_parameters_from_dataframe, _build_tcsc_ppc, _build_branch_dc_ppc
from pandapower.build_branch import _switch_branches, _branches_with_oos_buses, _initialize_branch_lookup, _end_temperature_correction_factor
from pandapower.pd2ppc import _ppc2ppci, _init_ppc


def _pd2ppc_zero(net, k_st, sequence=0):
    """
    Builds the ppc data structure for zero impedance system. Includes the impedance values of
    lines and transformers, but no load or generation data.

    For short-circuit calculation, the short-circuit impedance of external grids is also considered.
    """
    # select elements in service (time consuming, so we do it once)
    net["_is_elements"] = _select_is_elements_numba(net, sequence=sequence)

    ppc = _init_ppc(net, sequence)

    _build_bus_ppc(net, ppc)
    _build_bus_dc_ppc(net, ppc)
    _build_gen_ppc(net, ppc)
    _build_svc_ppc(net, ppc, "sc")   # needed for shape reasons
    _build_tcsc_ppc(net, ppc, "sc")  # needed for shape reasons
    _build_ssc_ppc(net, ppc, "sc")  # needed for shape reasons
    _build_vsc_ppc(net, ppc, "sc")  # needed for shape reasons
    _add_gen_sc_impedance_zero(net, ppc)
    _add_ext_grid_sc_impedance_zero(net, ppc)
    _build_branch_ppc_zero(net, ppc, k_st)
    _build_branch_dc_ppc(net, ppc)  # needed for shape reasons

    # adds auxilary buses for open switches at branches
    _switch_branches(net, ppc)

    # add auxilary buses for out of service buses at in service lines.
    # Also sets lines out of service if they are connected to two out of service buses
    _branches_with_oos_buses(net, ppc)
    if hasattr(net, "_isolated_buses"):
        ppc["bus"][net._isolated_buses, BUS_TYPE] = NONE
    if hasattr(net, "_isolated_buses_dc"):
        ppc["bus_dc"][net._isolated_buses_dc, DC_BUS_TYPE] = DC_NONE

    # generates "internal" ppci format (for powerflow calc) from "external" ppc format and updates the bus lookup
    # Note: Also reorders buses and gens in ppc
    ppci = _ppc2ppci(ppc, net)
    # net._ppc0 = ppc    <--Obsolete. now covered in _init_ppc
    return ppc, ppci


def _build_branch_ppc_zero(net, ppc, k_st=None):
    """
    Takes the empty ppc network and fills it with the zero imepdance branch values. The branch
    datatype will be np.complex 128 afterwards.

    .. note:: The order of branches in the ppc is:
            1. Lines
            2. Transformers

    **INPUT**:
        **net** -The pandapower format network

        **ppc** - The PYPOWER format network to fill in values

    """
    length = _initialize_branch_lookup(net)
    lookup = net._pd2ppc_lookups["branch"]
    mode = net._options["mode"]
    ppc["branch"] = np.zeros(shape=(length, branch_cols), dtype=np.complex128)
    if mode == "sc":
        branch_sc = np.empty(shape=(length, branch_cols_sc), dtype=float)
        branch_sc.fill(np.nan)
        ppc["branch"] = np.hstack((ppc["branch"], branch_sc))
    ppc["branch"][:, :13] = np.array([0, 0, 0, 0, 0, 250, 250, 250, 1, 0, 1, -360, 360])

    _add_line_sc_impedance_zero(net, ppc)
    _add_impedance_sc_impedance_zero(net, ppc)
    _add_trafo_sc_impedance_zero(net, ppc, k_st=k_st)
    if "switch" in lookup:
        _calc_switch_parameter(net, ppc)
    if mode == "sc":
        _add_trafo3w_sc_impedance_zero(net, ppc)
    else:
        if "trafo3w" in lookup:
            raise NotImplementedError("Three winding transformers are not implemented for unbalanced calculations")


def _add_trafo_sc_impedance_zero(net, ppc, trafo_df=None, k_st=None):
    if trafo_df is None:
        trafo_df = net["trafo"]
    if k_st is None:
        k_st = np.ones(len(ppc['branch']))
    if "xn_ohm" not in trafo_df.columns:
        trafo_df["xn_ohm"] = 0.
    branch_lookup = net["_pd2ppc_lookups"]["branch"]
    if "trafo" not in branch_lookup:
        return
    bus_lookup = net["_pd2ppc_lookups"]["bus"]
    mode = net["_options"]["mode"]
    trafo_model = net["_options"]["trafo_model"]
    f, t = branch_lookup["trafo"]
    trafo_df["_ppc_idx"] = range(f, t)
    trafo_df['k_st'] = k_st[trafo_df["_ppc_idx"].values].real
    bus_lookup = net["_pd2ppc_lookups"]["bus"]

    hv_bus = get_trafo_values(trafo_df, "hv_bus").astype(np.int64)
    lv_bus = get_trafo_values(trafo_df, "lv_bus").astype(np.int64)
    in_service = get_trafo_values(trafo_df, "in_service").astype(np.int64)
    ppc["branch"][f:t, F_BUS] = bus_lookup[hv_bus]
    ppc["branch"][f:t, T_BUS] = bus_lookup[lv_bus]
    buses_all, gs_all, bs_all = np.array([], dtype=np.int64), np.array([]), np.array([])
    BIG_NUMBER = 1e20 * ppc["baseMVA"]
    if mode == "sc":
        # Should be considered as connected for all in_service branches
        ppc["branch"][f:t, BR_X] = BIG_NUMBER
        ppc["branch"][f:t, BR_R] = BIG_NUMBER
        ppc["branch"][f:t, BR_B] = 0
        ppc["branch"][f:t, BR_STATUS] = in_service
    else:
        ppc["branch"][f:t, BR_STATUS] = 0

    if "vector_group" not in trafo_df:
        raise ValueError("Vector Group of transformer needs to be specified for zero "
                         "sequence modelling \n Try : net.trafo[\"vector_group\"] = 'Dyn'")

    for vector_group, trafos in trafo_df.groupby("vector_group"):
        # TODO Roman: check this/expand this
        ppc_idx = trafos["_ppc_idx"].values.astype(np.int64)

        if vector_group.lower() in ["yy", "yd", "dy", "dd"]:
            continue

        vk_percent = trafos["vk_percent"].values.astype(float)
        vkr_percent = trafos["vkr_percent"].values.astype(float)
        sn_trafo_mva = trafos["sn_mva"].values.astype(float)
        # Just put pos seq parameter if zero seq parameter is zero
        if "vk0_percent" not in trafos:
            raise ValueError("Short circuit voltage of transformer Vk0 needs to be specified for zero "
                             "sequence modelling \n Try : net.trafo[\"vk0_percent\"] = net.trafo[\"vk_percent\"]")
        vk0_percent = trafos["vk0_percent"].values.astype(float) if \
            trafos["vk0_percent"].values.astype(float).all() != 0. else \
            trafos["vk_percent"].values.astype(float)
        # Just put pos seq parameter if zero seq parameter is zero
        if "vkr0_percent" not in trafos:
            raise ValueError("Real part of short circuit voltage Vk0(Real) needs to be specified for transformer "
                             "modelling \n Try : net.trafo[\"vkr0_percent\"] = net.trafo[\"vkr_percent\"]")
        vkr0_percent = trafos["vkr0_percent"].values.astype(float) if \
            trafos["vkr0_percent"].values.astype(float).all() != 0. else \
            trafos["vkr_percent"].values.astype(float)
        lv_buses = trafos["lv_bus"].values.astype(np.int64)
        hv_buses = trafos["hv_bus"].values.astype(np.int64)
        lv_buses_ppc = bus_lookup[lv_buses]
        hv_buses_ppc = bus_lookup[hv_buses]
        if "mag0_percent" not in trafos:
            # For Shell Type transformers vk0 = vk * 1
            #        and mag0_percent = 10 ... 100  Zm0/ Zsc0
            # --pg 50 DigSilent Power Factory Transformer manual
            raise ValueError("Magnetizing impedance to vk0 ratio needs to be specified for transformer "
                             "modelling  \n Try : net.trafo[\"mag0_percent\"] = 100")
        mag0_ratio = trafos.mag0_percent.values.astype(float)
        if "mag0_rx" not in trafos:
            raise ValueError("Magnetizing impedance R/X ratio needs to be specified for transformer "
                             "modelling \n Try : net.trafo[\"mag0_rx\"] = 0 ")
        mag0_rx = trafos["mag0_rx"].values.astype(float)
        if "si0_hv_partial" not in trafos:
            raise ValueError("Zero sequence short circuit impedance partition towards HV side needs to be specified "
                             "for transformer modelling \n Try : net.trafo[\"si0_hv_partial\"] = 0.9 ")
        si0_hv_partial = trafos.si0_hv_partial.values.astype(float)
        parallel = trafos.parallel.values.astype(float)
        if "power_station_unit" in trafos.columns:
            power_station_unit = trafos.power_station_unit.fillna(False).astype(bool)
        else:
            power_station_unit = np.zeros(len(trafos), dtype=bool)
        in_service = trafos["in_service"].astype(np.int64)

        ppc["branch"][ppc_idx, F_BUS] = hv_buses_ppc
        ppc["branch"][ppc_idx, T_BUS] = lv_buses_ppc

        vn_trafo_hv, vn_trafo_lv, shift = _calc_tap_from_dataframe(net, trafos)
        vn_bus_lv = ppc["bus"][lv_buses_ppc, BASE_KV]
        vn_bus_hv = ppc["bus"][hv_buses_ppc, BASE_KV]
        ratio = _calc_nominal_ratio_from_dataframe(ppc, trafos, vn_trafo_hv,
                                                   vn_trafo_lv, bus_lookup)
        ppc["branch"][ppc_idx, TAP] = ratio
        ppc["branch"][ppc_idx, SHIFT] = shift

        # zero seq. transformer impedance
        tap_lv = np.square(vn_trafo_lv / vn_bus_lv) * net.sn_mva
        tap_hv = np.square(vn_trafo_hv / vn_bus_hv) * net.sn_mva
        if mode == 'pf_3ph':
            if vector_group.lower() not in ["ynyn", "dyn", "yzn"]:
                raise NotImplementedError("Calculation of 3-phase power flow is only implemented for the transformer "
                                          "vector groups 'YNyn', 'Dyn', 'Yzn'")
            # =============================================================================
            #     Changing base from transformer base to Network base to get Zpu(Net)
            #     Zbase = (kV).squared/S_mva
            #     Zpu(Net)={Zpu(trafo) * Zb(trafo)} / {Zb(Net)}
            #        Note:
            #             Network base voltage is Line-Neutral voltage in each phase
            #             Line-Neutral voltage= Line-Line Voltage(vn_bus_lv) divided by sq.root(3)
            # =============================================================================
            tap_lv = np.square(vn_trafo_lv / vn_bus_lv) * (3 * net.sn_mva)
            tap_hv = np.square(vn_trafo_hv / vn_bus_hv) * (3 * net.sn_mva)

        tap_corr = tap_hv if vector_group.lower() in ("ynd", "yny") else tap_lv
        z_sc = vk0_percent / 100. / sn_trafo_mva * tap_corr
        r_sc = vkr0_percent / 100. / sn_trafo_mva * tap_corr
        z_sc = z_sc.astype(float)
        r_sc = r_sc.astype(float)
        x_sc = np.sign(z_sc) * np.sqrt(z_sc ** 2 - r_sc ** 2)
        # TODO: This equation needs to be checked!
        # z0_k = (r_sc + x_sc * 1j) / parallel  * max(1, ratio) **2
        # z0_k = (r_sc + x_sc * 1j) / parallel * vn_trafo_hv / vn_bus_hv
        # z0_k = (r_sc + x_sc * 1j) / parallel * tap_hv
        z0_k = (r_sc + x_sc * 1j) / parallel
        z_n_ohm = trafos["xn_ohm"].fillna(0).values
        k_st_tr = trafos["k_st"].fillna(1).values

        if mode == "sc":  # or trafo_model == "pi":
            cmax = net._ppc["bus"][lv_buses_ppc, C_MAX]
            kt = _transformer_correction_factor(trafos, vk_percent, vkr_percent, sn_trafo_mva, cmax)
            z0_k *= kt

            # different formula must be applied for power station unit transformers:
            # z_0THV is for power station block unit transformer -> page 20 of IEC60909-4:2021 (example 4.4.2):
            # todo: check if sn_mva must be included here?
            vkx0_percent = np.sqrt(np.square(vk0_percent) - np.square(vkr0_percent))
            z_0THV = (vkr0_percent / 100 + 1j * vkx0_percent / 100) * (np.square(vn_trafo_hv) / sn_trafo_mva) / parallel
            z0_k_psu = (z_0THV * k_st_tr + 3j * z_n_ohm) / ((vn_bus_hv ** 2) / net.sn_mva)
            z0_k = np.where(power_station_unit, z0_k_psu, z0_k)

        y0_k = 1 / z0_k  # adding admittance for "pi" model
        # y0_k = 1 / (z0_k * k_st_tr + 3j * z_n_ohm)  # adding admittance for "pi" model

        # =============================================================================
        #       Transformer magnetising impedance for zero sequence
        # =============================================================================
        z_m = z_sc * mag0_ratio
        x_m = z_m / np.sqrt(mag0_rx ** 2 + 1)
        r_m = x_m * mag0_rx
        r0_trafo_mag = r_m / parallel
        x0_trafo_mag = x_m / parallel
        z0_mag = r0_trafo_mag + x0_trafo_mag * 1j
        # =============================================================================
        #         Star - Delta conversion ( T model to Pi Model)
        #      ----------- |__zc=ZAB__|-----------------
        #            _|                   _|
        #     za=ZAN|_|                  |_| zb=ZBN
        #            |                    |
        # =============================================================================
        z1 = si0_hv_partial * z0_k
        z2 = (1 - si0_hv_partial) * z0_k
        z3 = z0_mag
        z_temp = z1 * z2 + z2 * z3 + z1 * z3
        za = z_temp / z2
#        za = z_temp / (z2+z3)
        zb = z_temp / z1
#        zb = z_temp / (z1+z3)
        zc = z_temp / z3  # ZAB  Transfer impedance
#        zc = z_temp / (z1+z2)  # ZAB  Transfer impedance
        YAB = 1 / zc.astype(complex)
        YAN = 1 / za.astype(complex)
        YBN = 1 / zb.astype(complex)

#        YAB_AN = (zc + za) /(zc * za).astype(complex)  # Series conn YAB and YAN
#        YAB_BN = (zc + zb) / (zc * zb).astype(complex)  # Series conn YAB and YBN

        YAB_AN = 1 / (zc + za).astype(complex)  # Series conn YAB and YAN
        YAB_BN = 1 / (zc + zb).astype(complex)  # Series conn YAB and YBN

        # y0_k = 1 / z0_k #adding admittance for "pi" model
        if vector_group.lower() == "dyn":
            buses_all = np.hstack([buses_all, lv_buses_ppc])
            if trafo_model == "pi":
                y = y0_k * ppc["baseMVA"]  # pi model
            else:
                y = (YAB + YBN).astype(complex) * ppc["baseMVA"]  # T model
            gs_all = np.hstack([gs_all, y.real * in_service])
            bs_all = np.hstack([bs_all, y.imag * in_service])

        elif vector_group.lower() == "ynd":
            buses_all = np.hstack([buses_all, hv_buses_ppc])
            if trafo_model == "pi":
                y = y0_k * ppc["baseMVA"]  # pi model
                # y = 1/0.99598 * 1 / (1/(y0_k * ppc["baseMVA"]) + 1/0.99598 * (1j * 3 * 22 /( (110 ** 2) / 1))) # pi
                # y = 1/0.99598 * 1 / (1/(y0_k * ppc["baseMVA"]) + 1/0.99598 * (1j * 3 * 22 /( (110 ** 2) / 1))) # pi

                # z0_k_k = z0_k * 0.99598 + 1j * 3 * 22 /( (110 ** 2) / 1)
                # print(z0_k_k)
                # y = 1 / z0_k_k # pi model
            else:
                y = (YAB_BN + YAN).astype(complex) * ppc["baseMVA"]  # T model
            gs_all = np.hstack([gs_all, y.real * in_service])
            bs_all = np.hstack([bs_all, y.imag * in_service])

        elif vector_group.lower() == "yyn":
            buses_all = np.hstack([buses_all, lv_buses_ppc])
            if trafo_model == "pi":
                y = 1/(z0_mag+z0_k).astype(complex) * ppc["baseMVA"]  # pi model
            else:
                # y = (YAB_AN + YBN).astype(complex)  # T model
                y = (YAB + YAB_BN + YBN).astype(complex) * ppc["baseMVA"]  # T model

            gs_all = np.hstack([gs_all, y.real * in_service])
            bs_all = np.hstack([bs_all, y.imag * in_service])

        elif vector_group.lower() == "ynyn":
            ppc["branch"][ppc_idx, BR_STATUS] = in_service
            # zc = ZAB
            ppc["branch"][ppc_idx, BR_R] = zc.real
            ppc["branch"][ppc_idx, BR_X] = zc.imag

            buses_all = np.hstack([buses_all, hv_buses_ppc])
            gs_all = np.hstack([gs_all, YAN.real * in_service * ppc["baseMVA"] * tap_lv / tap_hv])
            bs_all = np.hstack([bs_all, YAN.imag * in_service * ppc["baseMVA"] * tap_lv / tap_hv])

            buses_all = np.hstack([buses_all, lv_buses_ppc])
            gs_all = np.hstack([gs_all, YBN.real * in_service * ppc["baseMVA"]])
            bs_all = np.hstack([bs_all, YBN.imag * in_service * ppc["baseMVA"]])

        elif vector_group.lower() == "yny":
            buses_all = np.hstack([buses_all, hv_buses_ppc])
            if trafo_model == "pi":
                y = 1/(z0_mag+z0_k).astype(complex) * ppc["baseMVA"]  # pi model
            else:
                y = (YAB_BN + YAN).astype(complex) * ppc["baseMVA"]  # T model
            gs_all = np.hstack([gs_all, y.real * in_service])
            bs_all = np.hstack([bs_all, y.imag * in_service])

        elif vector_group.lower() == "yzn":
            buses_all = np.hstack([buses_all, lv_buses_ppc])
            #            y = 1/(z0_mag+z0_k).astype(complex)* int(ppc["baseMVA"])#T model
            #            y= (za+zb+zc)/((za+zc)*zb).astype(complex)* int(ppc["baseMVA"])#pi model
            y = (YAB_AN + YBN).astype(complex) * ppc["baseMVA"] ** 2  # T model # why sn_mva squared here?
            gs_all = np.hstack([gs_all, (1.1547) * y.real * in_service])  # what's the 1.1547 value?
            bs_all = np.hstack([bs_all, (1.1547) * y.imag * in_service])

        elif vector_group[-1].isdigit():
            raise ValueError("Unknown transformer vector group %s - "
                             "please specify vector group without "
                             "phase shift number. Phase shift can be "
                             "specified in net.trafo.shift_degree" % vector_group)
        else:
            raise ValueError("Transformer vector group %s is unknown "
                             "/ not implemented for three phase load flow" % vector_group)

    buses, gs, bs = _sum_by_group(buses_all, gs_all, bs_all)
    ppc["bus"][buses, GS] += gs
    ppc["bus"][buses, BS] += bs
    del net.trafo["_ppc_idx"]


def _add_gen_sc_impedance_zero(net, ppc):
    mode = net["_options"]["mode"]
    if mode == 'pf_3ph':
        return

    eg = net["gen"][net._is_elements["gen"]]
    if len(eg) == 0:
        return
    eg_buses = eg.bus.values
    bus_lookup = net["_pd2ppc_lookups"]["bus"]
    eg_buses_ppc = bus_lookup[eg_buses]

    y0_gen = 1 / (1e3 + 1e3*1j)
    # buses, gs, bs = aux._sum_by_group(eg_buses_ppc, y0_gen.real, y0_gen.imag)
    ppc["bus"][eg_buses_ppc, GS] += y0_gen.real
    ppc["bus"][eg_buses_ppc, BS] += y0_gen.imag


def _add_ext_grid_sc_impedance_zero(net, ppc):
    mode = net["_options"]["mode"]

    if mode == "sc":
        case = net._options["case"]
    else:
        case = "max"
    bus_lookup = net["_pd2ppc_lookups"]["bus"]
    is_egs = net._is_elements["ext_grid"]
    eg = net["ext_grid"][is_egs]
    if len(eg) == 0:
        return
    eg_buses = eg.bus.values
    eg_buses_ppc = bus_lookup[eg_buses]

    if mode == "sc":
        c = ppc["bus"][eg_buses_ppc, C_MAX] if case == "max" else ppc["bus"][eg_buses_ppc, C_MIN]
    else:
        c = 1.1
    if not "s_sc_%s_mva" % case in eg:
        raise ValueError("short circuit apparent power s_sc_%s_mva needs to be specified for " % case +
                         "external grid")
    s_sc = eg["s_sc_%s_mva" % case].values
    if not "rx_%s" % case in eg:
        raise ValueError("short circuit R/X rate rx_%s needs to be specified for external grid" %
                         case)
    rx = eg["rx_%s" % case].values

    z_grid = c / s_sc
    if mode == 'pf_3ph':
        z_grid = c / (s_sc/3)
    x_grid = z_grid / np.sqrt(rx ** 2 + 1)
    r_grid = rx * x_grid
    eg["r"] = r_grid
    eg["x"] = x_grid

    # ext_grid zero sequence impedance
    x0_grid = net.ext_grid[is_egs]["x0x_%s" % case].values * x_grid
    r0_grid = net.ext_grid[is_egs]["r0x0_%s" % case].values * x0_grid
    y0_grid = 1 / (r0_grid + x0_grid*1j)

    buses, gs, bs = _sum_by_group(eg_buses_ppc, y0_grid.real, y0_grid.imag)
    ppc["bus"][buses, GS] = gs
    ppc["bus"][buses, BS] = bs


def _add_line_sc_impedance_zero(net, ppc):
    branch_lookup = net["_pd2ppc_lookups"]["branch"]
    mode = net["_options"]["mode"]
    if "line" not in branch_lookup:
        return
    line = net["line"]
    bus_lookup = net["_pd2ppc_lookups"]["bus"]
    length = line["length_km"].values
    parallel = line["parallel"].values

    fb = bus_lookup[line["from_bus"].values]
    tb = bus_lookup[line["to_bus"].values]
    baseR = np.square(ppc["bus"][fb, BASE_KV]) / net.sn_mva
    if mode == 'pf_3ph':
        baseR = np.square(ppc["bus"][fb, BASE_KV]) / (3*net.sn_mva)
    f, t = branch_lookup["line"]
    # line zero sequence impedance
    ppc["branch"][f:t, F_BUS] = fb
    ppc["branch"][f:t, T_BUS] = tb
    ppc["branch"][f:t, BR_R] = line["r0_ohm_per_km"].values * length / baseR / parallel
    if mode == "sc":
        # temperature correction
        if net["_options"]["case"] == "min":
            ppc["branch"][f:t, BR_R] *= _end_temperature_correction_factor(net, short_circuit=True)
    ppc["branch"][f:t, BR_X] = line["x0_ohm_per_km"].values * length / baseR / parallel
    ppc["branch"][f:t, BR_B] = (2 * net["f_hz"] * math.pi * line["c0_nf_per_km"].values *
                                1e-9 * baseR * length * parallel)
    ppc["branch"][f:t, BR_STATUS] = line["in_service"].astype(np.int64)


def _add_impedance_sc_impedance_zero(net, ppc):
    branch_lookup = net["_pd2ppc_lookups"]["branch"]
    if "impedance" not in branch_lookup:
        return
    bus_lookup = net["_pd2ppc_lookups"]["bus"]
    branch = ppc["branch"]

    f, t = branch_lookup["impedance"]

    # impedance zero sequence impedance
    rij, xij, r_asym, x_asym, gi, bi, g_asym, b_asym = _calc_impedance_parameters_from_dataframe(net, zero_sequence=True)
    branch[f:t, BR_R] = rij
    branch[f:t, BR_X] = xij
    branch[f:t, BR_R_ASYM] = r_asym
    branch[f:t, BR_X_ASYM] = x_asym
    branch[f:t, BR_G] = gi
    branch[f:t, BR_B] = bi
    branch[f:t, BR_G_ASYM] = g_asym
    branch[f:t, BR_B_ASYM] = b_asym
    branch[f:t, F_BUS] = bus_lookup[net.impedance["from_bus"].values]
    branch[f:t, T_BUS] = bus_lookup[net.impedance["to_bus"].values]
    branch[f:t, BR_STATUS] = net["impedance"]["in_service"].values.astype(np.int64)


def _add_trafo3w_sc_impedance_zero(net, ppc):
    # TODO Roman: check this/expand this
    branch_lookup = net["_pd2ppc_lookups"]["branch"]
    if "trafo3w" not in branch_lookup:
        return
    bus_lookup = net["_pd2ppc_lookups"]["bus"]
    branch = ppc["branch"]
    f, t = net["_pd2ppc_lookups"]["branch"]["trafo3w"]
    trafo_df = _trafo_df_from_trafo3w(net, sequence=0)
    hv_bus = get_trafo_values(trafo_df, "hv_bus").astype(np.int64)
    lv_bus = get_trafo_values(trafo_df, "lv_bus").astype(np.int64)
    in_service = get_trafo_values(trafo_df, "in_service").astype(np.int64)
    branch[f:t, F_BUS] = bus_lookup[hv_bus]
    branch[f:t, T_BUS] = bus_lookup[lv_bus]

    r, x, *_, ratio, shift = _calc_branch_values_from_trafo_df(net, ppc, trafo_df, sequence=0)

    # Y0y0d5,  YN0y0d5,  Y0yn0d5,  YN0yn0d5, Y0y0y0, Y0d5d5,
    # YN0d5d5,  Y0d5y0,  Y0y0d11  und  D0d0d0

    #
    # already implemented:
    #   YNyd
    #   YYnd
    #   YNynd
    #   YNdd
    # not relevant (for 1ph):
    #   Yyy
    #   Ydd
    #   Ydy
    #   Yyd
    #   Ddd

    BIG_NUMBER = 1e20 * ppc["baseMVA"]

    n_t3 = net.trafo3w.shape[0]
    for t3_ix in np.arange(n_t3):
        t3 = net.trafo3w.iloc[t3_ix, :]

        if t3.vector_group.lower() in set(map(lambda vg: "".join(vg), product("dy", repeat=3))):
            x[[t3_ix, t3_ix+n_t3, t3_ix+n_t3*2]] = BIG_NUMBER
            r[[t3_ix, t3_ix+n_t3, t3_ix+n_t3*2]] = BIG_NUMBER
        elif t3.vector_group.lower() == "ynyd":
            # Correction for YNyd
            # z3->y3
            ys = ppc["baseMVA"] / ((x[t3_ix+n_t3*2] * 1j + r[t3_ix+n_t3*2]) * ratio[t3_ix+n_t3*2] ** 2)
            aux_bus = bus_lookup[lv_bus[t3_ix]]
            ppc["bus"][aux_bus, BS] += ys.imag
            ppc["bus"][aux_bus, GS] += ys.real

            # Set y2/y3 to almost 0 to avoid isolated bus
            x[[t3_ix+n_t3, t3_ix+n_t3*2]] = BIG_NUMBER
            r[[t3_ix+n_t3, t3_ix+n_t3*2]] = BIG_NUMBER
        elif t3.vector_group.lower() == "yndy":
            # Correction for YNyd
            # z3->y3
            ys = ppc["baseMVA"] / ((x[t3_ix + n_t3] * 1j + r[t3_ix + n_t3]) * ratio[t3_ix + n_t3] ** 2)
            aux_bus = bus_lookup[lv_bus[t3_ix]]
            ppc["bus"][aux_bus, BS] += ys.imag
            ppc["bus"][aux_bus, GS] += ys.real

            # Set y2/y3 to almost 0 to avoid isolated bus
            x[[t3_ix+n_t3, t3_ix+n_t3*2]] = BIG_NUMBER
            r[[t3_ix+n_t3, t3_ix+n_t3*2]] = BIG_NUMBER
        elif t3.vector_group.lower() == "yynd":
            # z3->y3
            ys = ppc["baseMVA"] / ((x[t3_ix+n_t3*2] * 1j + r[t3_ix+n_t3*2]) * ratio[t3_ix+n_t3*2] ** 2)
            aux_bus = bus_lookup[lv_bus[t3_ix]]
            ppc["bus"][aux_bus, BS] += ys.imag
            ppc["bus"][aux_bus, GS] += ys.real

            # Set y1/y3 to almost 0 to avoid isolated bus
            x[[t3_ix, t3_ix+n_t3*2]] = BIG_NUMBER
            r[[t3_ix, t3_ix+n_t3*2]] = BIG_NUMBER
        elif t3.vector_group.lower() == "ydyn":
            # z3->y3
            ys = ppc["baseMVA"] / ((x[t3_ix+n_t3] * 1j + r[t3_ix+n_t3]) * ratio[t3_ix+n_t3] ** 2)
            aux_bus = bus_lookup[lv_bus[t3_ix]]
            ppc["bus"][aux_bus, BS] += ys.imag
            ppc["bus"][aux_bus, GS] += ys.real

            # Set y1/y3 to almost 0 to avoid isolated bus
            x[[t3_ix, t3_ix+n_t3]] = BIG_NUMBER
            r[[t3_ix, t3_ix+n_t3]] = BIG_NUMBER
        elif t3.vector_group.lower() == "ynynd":
            # z3->y3
            ys = ppc["baseMVA"] / ((x[t3_ix+n_t3*2] * 1j + r[t3_ix+n_t3*2]) * ratio[t3_ix+n_t3*2] ** 2)
            aux_bus = bus_lookup[lv_bus[t3_ix]]
            ppc["bus"][aux_bus, BS] += ys.imag
            ppc["bus"][aux_bus, GS] += ys.real

            # Set y3 to almost 0 to avoid isolated bus
            x[t3_ix+n_t3*2] = BIG_NUMBER
            r[t3_ix+n_t3*2] = BIG_NUMBER
        elif t3.vector_group.lower() == "yndyn":
            # z3->y3
            ys = ppc["baseMVA"] / ((x[t3_ix + n_t3] * 1j + r[t3_ix + n_t3]) * ratio[t3_ix + n_t3] ** 2)
            aux_bus = bus_lookup[lv_bus[t3_ix]]
            ppc["bus"][aux_bus, BS] += ys.imag
            ppc["bus"][aux_bus, GS] += ys.real

            # Set y3 to almost 0 to avoid isolated bus
            x[t3_ix + n_t3] = BIG_NUMBER
            r[t3_ix + n_t3] = BIG_NUMBER
        elif t3.vector_group.lower() == "yndd":
            # Correction for YNdd
            # z3->y3
            # we need a shunt impedance for both "delta" windings -> ys1, ys2
            ys1 = ppc["baseMVA"] / ((x[t3_ix + n_t3] * 1j + r[t3_ix + n_t3]) * ratio[t3_ix + n_t3] ** 2)
            ys2 = ppc["baseMVA"] / ((x[t3_ix + n_t3 * 2] * 1j + r[t3_ix + n_t3 * 2]) * ratio[t3_ix + n_t3 * 2] ** 2)
            aux_bus = bus_lookup[lv_bus[t3_ix]]
            ppc["bus"][aux_bus, BS] += ys1.imag + ys2.imag
            ppc["bus"][aux_bus, GS] += ys1.real + ys2.real

            # Set y2/y3 to almost 0 to avoid isolated bus
            x[[t3_ix + n_t3, t3_ix + n_t3 * 2]] = BIG_NUMBER
            r[[t3_ix + n_t3, t3_ix + n_t3 * 2]] = BIG_NUMBER
        elif t3.vector_group.lower() == "ynyy":
            # Correction for YNyy
            x[[t3_ix, t3_ix + n_t3, t3_ix + n_t3 * 2]] = BIG_NUMBER
            r[[t3_ix, t3_ix + n_t3, t3_ix + n_t3 * 2]] = BIG_NUMBER
        else:
            raise UserWarning(f"{t3.vector_group} not supported yet for trafo3w!")

    branch[f:t, BR_R] = r
    branch[f:t, BR_X] = x
    branch[f:t, BR_B] = 0
    branch[f:t, TAP] = ratio
    branch[f:t, SHIFT] = shift
    branch[f:t, BR_STATUS] = in_service
